from pya import *
import pya

class Wireguide(pya.PCellDeclarationHelper):
  '''
  RIT Integrated Photonics Group of Stefan Preble (Karl McNulty) 
  new wireguide class (essentially the metal path equivalent to a waveguide but no bend is allowed)
  '''
  def __init__(self):
    # Important: initialize the super class
    super(Wireguide, self).__init__()
    # declare the parameters
    from SiEPIC.utils import get_technology_by_name    
    TECHNOLOGY = get_technology_by_name('EBeam')
    self.param("path", self.TypeShape, "Path", default = DPath([DPoint(0,0), DPoint(10,0), DPoint(10,10)], 0.5))
    self.radius = 0
    self.param("width", self.TypeDouble, "Width", default = 0.5)
    self.adiab = False
    self.param("layers", self.TypeList, "Layers", default = ['M2_router'])
    self.param("widths", self.TypeList, "Widths", default =  [5])
    self.param("offsets", self.TypeList, "Offsets", default = [0])
   
  def display_text_impl(self):
    # Provide a descriptive text for the cell
    return "Wireguide_%s" % self.path
  
  def coerce_parameters_impl(self):
    return
          
  def can_create_from_shape_impl(self):
    return self.shape.is_path()

  def transformation_from_shape_impl(self):
    return Trans(Trans.R0,0,0)

  def parameters_from_shape_impl(self):
    self.path = self.shape.dpath
        
  def produce_impl(self):

    from SiEPIC.utils import arc_xy, arc_bezier, angle_vector, angle_b_vectors, inner_angle_b_vectors, translate_from_normal
    from math import cos, sin, pi, sqrt
    import pya
    from SiEPIC.extend import to_itype
    
    # print("EBeam.Wireguide")
    
    from SiEPIC.utils import get_technology_by_name
    TECHNOLOGY = get_technology_by_name('EBeam')
    
    dbu = self.layout.dbu
    wg_width = to_itype(self.width,dbu)
    path = self.path.to_itype(dbu)

    if not (len(self.layers)==len(self.widths) and len(self.layers)==len(self.offsets) and len(self.offsets)==len(self.widths)):
      raise Exception("There must be an equal number of layers, widths and offsets")
    path.unique_points()
    turn=0
    for lr in range(0, len(self.layers)):
      layer = self.layout.layer(TECHNOLOGY[self.layers[lr]])
      
      width = to_itype(self.widths[lr],dbu)
      offset = to_itype(self.offsets[lr],dbu)

      pts = path.get_points()
      wg_pts = [pts[0]]
      for i in range(1,len(pts)-1):
        turn = ((angle_b_vectors(pts[i]-pts[i-1],pts[i+1]-pts[i])+90)%360-90)/90
        dis1 = pts[i].distance(pts[i-1])
        dis2 = pts[i].distance(pts[i+1])
        angle = angle_vector(pts[i]-pts[i-1])/90
        pt_radius = self.radius
        # determine the radius, based on how much space is available
        if len(pts)==3:
          pt_radius = min (dis1, dis2, pt_radius)
        else:
          if i==1:
            if dis1 <= pt_radius:
              pt_radius = dis1
          elif dis1 < 2*pt_radius:
            pt_radius = dis1/2
          if i==len(pts)-2:
            if dis2 <= pt_radius:
              pt_radius = dis2
          elif dis2 < 2*pt_radius:
            pt_radius = dis2/2
        # wireguide bends:
        if(self.adiab):
          wg_pts += Path(arc_bezier(pt_radius, 270, 270 + inner_angle_b_vectors(pts[i-1]-pts[i], pts[i+1]-pts[i]), self.bezier, DevRec='DevRec' in self.layers[lr], dbu=dbu), 0).transformed(Trans(angle, turn < 0, pts[i])).get_points()
        else:
          wg_pts += Path(arc_xy(-pt_radius, pt_radius, pt_radius, 270, 270 + inner_angle_b_vectors(pts[i-1]-pts[i], pts[i+1]-pts[i]),DevRec='DevRec' in self.layers[lr], dbu=dbu), 0).transformed(Trans(angle, turn < 0, pts[i])).get_points()
      wg_pts += [pts[-1]]
      wg_pts = pya.Path(wg_pts, 0).unique_points().get_points()
      wg_polygon = Path(wg_pts, wg_width)
      self.cell.shapes(layer).insert(wg_polygon) # insert the wireguide
       
      #if self.layout.layer(TECHNOLOGY['Wireguide']) == layer:
      waveguide_length = wg_polygon.area() / self.width * dbu**2

    pts = path.get_points()
    LayerPinRecN = self.layout.layer(TECHNOLOGY['PinRecM'])
    
    # insert pins to wireguide
    t1 = Trans(angle_vector(pts[0]-pts[1])/90, False, pts[0])
    self.cell.shapes(LayerPinRecN).insert(Path([Point(-50, 0), Point(50, 0)], wg_width).transformed(t1))
    self.cell.shapes(LayerPinRecN).insert(Text("pin1", t1, 0.3/dbu, -1))
    
    t = Trans(angle_vector(pts[-1]-pts[-2])/90, False, pts[-1])
    self.cell.shapes(LayerPinRecN).insert(Path([Point(-50, 0), Point(50, 0)], wg_width).transformed(t))
    self.cell.shapes(LayerPinRecN).insert(Text("pin2", t, 0.3/dbu, -1))
	
    LayerDevRecN = self.layout.layer(TECHNOLOGY['DevRec'])
    
    # Compact model information
    angle_vec = angle_vector(pts[0]-pts[1])/90
    halign = 0 # left
    angle=0
    pt2=pts[0]
    pt3=pts[0]
    if angle_vec == 0: # horizontal
      halign = 2 # right
      angle=0
      pt2=pts[0] + Point(0, wg_width)
      pt3=pts[0] + Point(0, -wg_width)
    if angle_vec == 2: # horizontal
      halign = 0 # left
      angle = 0
      pt2=pts[0] + Point(0, wg_width)
      pt3=pts[0] + Point(0, -wg_width)
    if angle_vec == 1: # vertical
      halign = 2 # right
      angle = 1
      pt2=pts[0] + Point(wg_width,0)
      pt3=pts[0] + Point(-wg_width,0)
    if angle_vec == -1: # vertical
      halign = 0 # left
      angle = 1
      pt2=pts[0] + Point(wg_width,0)
      pt3=pts[0] + Point(-wg_width,0)



print('SiEPIC-EBeam-PDK Python module: siepic_ebeam_pdk, KLayout technology: EBeam')

# Load the KLayout technology, when running in Script mode
import pya, os
if not pya.Technology().has_technology('EBeam'):
    tech = pya.Technology().create_technology('EBeam')
    tech.load(os.path.join(os.path.dirname(os.path.realpath(__file__)), 'EBeam.lyt'))

# then import all the technology modules
from . import pymacros





import pya
from pya import *
from SiEPIC.utils import get_technology_by_name
from . import *

class cavity_hole(pya.PCellDeclarationHelper):
  """
  The PCell declaration for the cavity hole.
  This is an example of how to perform inversion operations
  by Jaspreet, Jingda, and Lukas
  """

  def __init__(self):

    # Important: initialize the super class
    super(cavity_hole, self).__init__()
    TECHNOLOGY = get_technology_by_name('EBeam')

    # declare the parameters
    self.param("x", self.TypeDouble, "x coordinate", default = 0)
    self.param("y", self.TypeDouble, "y coordinate", default = 0)
    self.param("radius", self.TypeDouble, "hole radius", default = 10)
    self.param("gap", self.TypeDouble, "half of the gap between surrounding holes (microns)", default = 2)
    self.param("LayerSi", self.TypeLayer, "Si Layer", default = TECHNOLOGY['Si'])
    self.param("pinrec", self.TypeLayer, "PinRec Layer", default = TECHNOLOGY['PinRec'])
    self.param("devrec", self.TypeLayer, "DevRec Layer", default = TECHNOLOGY['DevRec'])
    

  def display_text_impl(self):
    # Provide a descriptive text for the cell
    return "cavity_hole(R=" + ('%.3f-%.3f' % (self.radius,self.gap) ) + ")"

  def can_create_from_shape(self, layout, shape, layer):
    return False


  def produce_impl(self):   
    import pya
    from operator import xor
    from math import pi, cos, sin
    
    dbu = self.layout.dbu
    ly = self.layout
    shapes = self.cell.shapes

    #Layers definitions
    LayerSi = self.layer
    LayerSiN = ly.layer(self.LayerSi)
    LayerPinRecN = ly.layer(self.pinrec)
    LayerDevRecN = ly.layer(self.devrec)
    TECHNOLOGY = get_technology_by_name('EBeam')
    LayerTextN = TECHNOLOGY['Text']
    # cell: layout cell to place the layout
    # LayerSiN: which layer to use
    # x, y: location of the origin
    # r: radius
    # w: waveguide width
    # length units in dbu
    
    x = self.x/dbu
    y = self.y/dbu
    r = self.radius/dbu
    r= r/2
    w = 2*r;
    
    gap = self.gap/dbu #gap between one hole to another

  
    # function to generate points to create a circle
    def circle(x,y,r):
      npts = 32
      theta = 2 * math.pi / npts # increment, in radians
      pts = []
      for i in range(0, npts):
        pts.append(Point.from_dpoint(DPoint((x+r*math.cos(i*theta))/1, (y+r*math.sin(i*theta))/1)))
      return pts
  
    
    #rectangular shapes
    side_gap = gap # half of gap between surrounding holes
    dx = side_gap + (2*r)
    dy = side_gap + (2*r)

    Si_slab = Region()
    Si_slab.insert(Box(-dx/2, -dy/2, dx/2, dy/2))
     
    # Circle 
    hole = Region()
    hole_cell = circle(0,0,r)
    hole_poly = Polygon(hole_cell)  
    hole_t = hole_poly.transformed(Trans(Trans.R0, x,y))
    hole.insert(hole_t)           

    # perform inversion:
    phc = Si_slab - hole
    self.cell.shapes(LayerSiN).insert(phc)
    # inversion can also be done by performing the XOR function
    # upper_poly = xor(Region(Polygon(circleu_pts)),Region(Polygon(rectu_pts)))

    
    return "cavity_hole(R=" + ('%.3f-%.3f' % (r,gap) ) + ")"



# $description: Unit testing: all EBeam libraries' cells
# $show-in-menu
# $group-name: Examples_EBeam
# $menu-path: siepic_menu.exlayout.end
# Unit testing for all library layout fixed cells and PCells

import pya # klayout
import os, sys

path_GitHub = os.path.expanduser('~/Documents/GitHub/')
if 0:
    if 'SiEPIC' not in sys.modules:
        path_siepic = os.path.join(path_GitHub, 'SiEPIC-Tools/klayout_dot_config/python')
        if not path_siepic in sys.path:
            sys.path.insert(0,path_siepic)
import SiEPIC
from SiEPIC._globals import Python_Env
print('KLayout running in mode: %s' % Python_Env)
from SiEPIC.utils.layout import new_layout, floorplan
from SiEPIC.utils import load_klayout_technology
from SiEPIC.scripts import instantiate_all_library_cells, zoom_out, export_layout

path_module = os.path.join(path_GitHub, 'SiEPIC_EBeam_PDK/klayout')
path_lyt_file = os.path.join(path_GitHub, 'SiEPIC_EBeam_PDK/klayout/EBeam/EBeam.lyt')
tech = load_klayout_technology('EBeam', path_module, path_lyt_file)


# Create a new layout
topcell, ly = new_layout(tech.name, "UnitTesting", GUI=True, overwrite = True)
# Instantiate all cells
instantiate_all_library_cells(topcell)
# Save the layout
path_out = os.path.dirname(os.path.realpath(__file__))
ly.write(os.path.join(path_out,'a.gds'))
export_layout(topcell, path_out,'a_static',format='oas')
print('done')


def get_error_layer_shapes(ly):
    '''Get all the shapes contained on the KLayout Layout.error_layer
    '''
    
#    print("layout: %s" % (ly))
    layer = ly.error_layer() # ly.get_info(ly.error_layer())

    for c in ly.cells('*'):
#        print("cell: %s" % (c.name))
        it = c.begin_shapes_rec(layer)
        while not(it.at_end()):
#            print(" shape: %s" % it.shape())
            if it.shape().is_text():
                print("error_layer: %s" % (it.shape().text.string))
            it.next()

get_error_layer_shapes(ly)

                
            



import pya
from pya import *
from SiEPIC.utils import get_technology_by_name

class Tapered_Ring(pya.PCellDeclarationHelper):
  """
  The PCell declaration for the tapered ring.
  Author: Mustafa Hammood     mustafa@ece.ubc.ca
  """

  def __init__(self):

    # Important: initialize the super class
    super(Tapered_Ring, self).__init__()
    TECHNOLOGY = get_technology_by_name('EBeam')

    # declare the parameters
    self.param("w_top", self.TypeDouble, "Top width", default = .5)
    self.param("w_bot", self.TypeDouble, "Bottom width (>Top Width)", default = 2)
    self.param("radius", self.TypeDouble, "Radius", default = 10)
    self.param("LayerSi", self.TypeLayer, "Si Layer", default = TECHNOLOGY['Si'])
    self.param("pinrec", self.TypeLayer, "PinRec Layer", default = TECHNOLOGY['PinRec'])
    self.param("devrec", self.TypeLayer, "DevRec Layer", default = TECHNOLOGY['DevRec'])


  def display_text_impl(self):
    # Provide a descriptive text for the cell
    return "Tapered_Ring(R=" + ('%.3f-%.3f' % (self.radius,self.w_bot) ) + ")"

  def can_create_from_shape(self, layout, shape, layer):
    return False


  def produce_impl(self):   
    import pya
    from operator import xor
    from math import pi, cos, sin
    from pya import Region, Polygon
    import math
    
    #fetch parameters
    dbu = self.layout.dbu
    ly = self.layout
    shapes = self.cell.shapes

    #Layers definitions
    LayerSi = self.layer
    LayerSiN = ly.layer(self.LayerSi)
    LayerPinRecN = ly.layer(self.pinrec)
    LayerDevRecN = ly.layer(self.devrec)
    #LayerTextN = ly.layer(LayerText)
    
    # cell: layout cell to place the layout
    # LayerSiN: which layer to use
    # x, y: location of the origin
    # r: radius
    # w: waveguide width
    # length units in dbu

    x = 0
    w_top = self.w_top/dbu
    r = self.radius/dbu    
   
    w_bot = self.w_bot/dbu
    
    
    deltaW = w_bot - w_top
    
    nptsFactor = 12
    
    
    # function to generate points to create circle
    def circle(x,w_top,r):
      npts = 32*nptsFactor
      theta =  2*math.pi / npts # increment, in radians
      pts = []
      for i in range(0, npts):
        pts.append(Point.from_dpoint(DPoint((x+r*math.cos(i*theta))/1, (w_top+r*math.sin(i*theta))/1)))
      return pts
  
    # function to generate points to create innercircle
    def inner_circle(x,w_top,r):
      npts = 32*nptsFactor
      theta =  2*math.pi / npts # increment, in radians
      pts = []
      for i in range(0, npts):
        pts.append(Point.from_dpoint(DPoint((x+r*math.cos(i*theta))/1, (w_top+r*math.sin(i*theta))/1)))
      return pts
     
  
    # Outer circle
    x = 0
    y = 0
    r_out = r+w_top/2
    ring = Region()
    ring_cell = circle(0,0,r_out)
    ring_poly = Polygon(ring_cell)  
    #ring_t = ring_poly.transformed(Trans(x,y))
    ring.insert(ring_poly)
    
    
    # Inner erasing circle  
    r_in = r - (w_top)/2 - deltaW/2
    x = 0
    y = -deltaW/2
    # Inner Circle 
    hole = Region()
    hole_cell = inner_circle(x,y,r_in)
    hole_poly = Polygon(hole_cell)  
    #hole_t = hole_poly.transformed(Trans(x,y))
    hole.insert(hole_poly)           
    
    # perform inversion:
    phc = ring - hole
    self.cell.shapes(LayerSiN).insert(phc)
    # inversion can also be done by performing the XOR function
    # upper_poly = xor(Region(Polygon(circleu_pts)),Region(Polygon(rectu_pts)))
    
    return "Tapered_Ring(R=" + ('%.3f-%.3f' % (r,w_bot) ) + ")"



import pya
from pya import *
from SiEPIC.utils import get_technology_by_name

class Waveguide_SWG_Assist(pya.PCellDeclarationHelper):
  """
  Input: length, target_period, wg_width, width, duty
  """

  def __init__(self):

    # Important: initialize the super class
    super(Waveguide_SWG_Assist, self).__init__()
    TECHNOLOGY = get_technology_by_name('EBeam')

    # declare the parameters
    self.param("length", self.TypeDouble, "Waveguide length", default = 10.0)     
    self.param("target_period", self.TypeDouble, "Target period (microns)", default = 0.200)     
    self.param("swg_wg_width", self.TypeDouble, "SWG Waveguide width", default = 0.5)     
    self.param("strip_wg_width", self.TypeDouble, "Strip Waveguide width", default = 0.06)     
    self.param("duty", self.TypeDouble, "Duty Cycle (0 to 1)", default = 0.5)
    self.param("layer", self.TypeLayer, "Layer", default = TECHNOLOGY['Si'])
    self.param("pinrec", self.TypeLayer, "PinRec Layer", default = TECHNOLOGY['PinRec'])
    self.param("devrec", self.TypeLayer, "DevRec Layer", default = TECHNOLOGY['DevRec'])
#    self.param("textl", self.TypeLayer, "Text Layer", default = LayerInfo(10, 0))

  def display_text_impl(self):
    # Provide a descriptive text for the cell
    return "Waveguide_SWG_Assist_%s-%.3f-%.3f-%.3f-%.3f" % \
    (self.length, self.target_period, self.swg_wg_width, self.strip_wg_width, self.duty)
  
  def coerce_parameters_impl(self):
    pass

  def can_create_from_shape(self, layout, shape, layer):
    return False
    
  def produce_impl(self):
  
    # fetch the parameters
    dbu = self.layout.dbu
    ly = self.layout
    shapes = self.cell.shapes

    LayerSi = self.layer
    LayerSiN = ly.layer(LayerSi)
    #LayerSiSPN = ly.layer(LayerSiSP)
    LayerPinRecN = ly.layer(self.pinrec)
    LayerDevRecN = ly.layer(self.devrec)

    # Determine the period such that the waveguide length is as desired.  Slight adjustment to period
    N_boxes = int(round(self.length / self.target_period-0.5))
    grating_period = self.length / (N_boxes) / dbu
    # print("N boxes: %s, grating_period: %s" % (N_boxes, grating_period) )
    
    # Draw the Bragg grating:
    box_width = int(round(grating_period*self.duty))
    
    w = self.swg_wg_width / dbu
    half_w = w/2
    for i in range(0,N_boxes+1):
      x = int(round((i * grating_period - box_width/2)))
      box1 = Box(x, -half_w, x + box_width, half_w)
      shapes(LayerSiN).insert(box1)
    length = self.length / dbu

    # Strip waveguide:
    w_strip = self.strip_wg_width / dbu
    if w_strip > 0:
        box1 = Box(0, -w_strip/2, length, w_strip/2)
        shapes(LayerSiN).insert(box1)

    # Pins on the waveguide:
    from SiEPIC._globals import PIN_LENGTH as pin_length
#    pin_length = box_width
    t = Trans(Trans.R0, 0,0)
    pin = Path([Point(pin_length/2, 0), Point(-pin_length/2, 0)], w)
    pin_t = pin.transformed(t)
    shapes(LayerPinRecN).insert(pin_t)
    text = Text ("pin1", t)
    shape = shapes(LayerPinRecN).insert(text)
    shape.text_size = 0.4/dbu

    t = Trans(Trans.R0, length,0)
    pin = Path([Point(-pin_length/2, 0), Point(pin_length/2, 0)], w)
    pin_t = pin.transformed(t)
    shapes(LayerPinRecN).insert(pin_t)
    text = Text ("pin2", t)
    shape = shapes(LayerPinRecN).insert(text)
    shape.text_halign = 2
    shape.text_size = 0.4/dbu

    # Compact model information
    t = Trans(Trans.R0, 0, 0)
    text = Text ('Lumerical_INTERCONNECT_library=Design kits/ebeam', t)
    shape = shapes(LayerDevRecN).insert(text)
    shape.text_size = 0.1/dbu
    t = Trans(Trans.R0, length/10, 0)
    text = Text ('Component=NO_MODEL_AVAILABLE', t)
    shape = shapes(LayerDevRecN).insert(text)
    shape.text_size = 0.1/dbu
    t = Trans(Trans.R0, length/9, -box_width*2)
    text = Text \
      ('Spice_param:length=%.3fu target_period=%.3fu grating_period=%.3fu swg_wg_width=%.3fu strip_wg_width=%.3fu duty=%.3f ' %\
      (self.length, self.target_period, round(grating_period)*dbu, self.swg_wg_width, self.strip_wg_width, self.duty), t )
    shape = shapes(LayerDevRecN).insert(text)
    shape.text_size = 0.1/dbu

    # Create the device recognition layer -- make it 1 * wg_width away from the waveguides.
    points = [pya.Point(0,0), pya.Point(length, 0)]
    path = pya.Path(points,w)
    path = pya.Path(points,w*3)
    shapes(LayerDevRecN).insert(path.simple_polygon())



import pya
from pya import *
from SiEPIC.utils import get_technology_by_name

class MultiBox_BraggGrating(pya.PCellDeclarationHelper):
    '''
    Reference:
    Enxiao Luan, Han Yun, Minglei Ma, Daniel M. Ratner, Karen C. Cheung, and Lukas Chrostowski,
    "Label-free biosensing with a multi-box sub-wavelength phase-shifted Bragg grating waveguide"
    Biomedical Optics Express Vol. 10, Issue 9, pp. 4825-4838 (2019)
    https://doi.org/10.1364/BOE.10.004825
    Multibox Bragg Grating
    By Enxiao Luan at UBC, 2020, with fix by Lukas Chrostowski
    '''

    def __init__(self):

      # Important: initialize the super class
      super(MultiBox_BraggGrating, self).__init__()
      TECHNOLOGY = get_technology_by_name('EBeam')

      # declare the parameters
      self.param("pitch", self.TypeDouble, "Period of SWG [um]", default = 0.24)
      self.param("w", self.TypeDouble, "Waveguide Width [um]", default = 0.18)
      self.param("wc", self.TypeDouble, "Corrugation width [um]", default = 0.1) # minimum radius is 5 um in TE
      self.param("ff", self.TypeDouble, "Duty Cycle", default = 0.75)
      #self.param("angle", self.TypeDouble, "Angle", default = 360)
      self.param("gap", self.TypeDouble, "Row Gap [um]", default = 0.06)
      #self.param("gap2", self.TypeDouble, "Bus Gap [um]", default = 0.06)
      self.param("row", self.TypeDouble, "Number of Rows of main WG", default = 3)
      self.param("Length", self.TypeDouble, "Bragg Grating Length on Each side [um]", default = 30)
      self.param("taperL", self.TypeDouble, "Taper Length [um]", default = 10)
      self.param("phaseshifted", self.TypeBoolean, "Bragg Type (Uniform=False, Phase-shifted=True)", default = False)
      self.param("layer", self.TypeLayer, "Layer", default = TECHNOLOGY['Si'])
      self.param("pinrec", self.TypeLayer, "PinRec Layer", default = TECHNOLOGY['PinRec'])
      self.param("devrec", self.TypeLayer, "DevRec Layer", default = TECHNOLOGY['DevRec'])
  #    self.param("textl", self.TypeLayer, "Text Layer", default = LayerInfo(10, 0))
  
    def display_text_impl(self):
      # Provide a descriptive text for the cell
      return "MultiBox_BraggGrating_length%.3f um " % \
      (2*self.Length)

    def coerce_parameters_impl(self):
       pass

    def can_create_from_shape(self, layout, shape, layer):
        return False
        
    def produce_impl(self):
      from SiEPIC._globals import PIN_LENGTH
      from SiEPIC.extend import to_itype
      from pya import DPolygon
      import math
      # This is the main part of the implementation: create the layout

      # fetch the parameters
      dbu = self.layout.dbu
      ly = self.layout
      shapes = self.cell.shapes
      LayerSi = self.layer
      LayerSiN = ly.layer(LayerSi)
      LayerPinRecN = ly.layer(self.pinrec)
      LayerDevRecN = ly.layer(self.devrec)
      
      #variables
      pitch = self.pitch
      w = self.w
      wc = self.wc
      ff = self.ff
      #angle = self.angle
      gap = self.gap
      #gap2 = self.gap2
      row = self.row
      Length = self.Length
      taperL = self.taperL
      phaseshifted = self.phaseshifted
      #print(doublebus)
      
      if Length <= 3*pitch:
        Length = 3*pitch
        print('invalid length of MultiBox Bragg grating, set length at least 3 times larger than the period of multibox blocks')
    
      s1 = pitch*ff #silicon
      s2 = pitch-s1 #gap


      # Draw the first Multi-box Waveguides 
      #calulate ideal length of bus
      Bragg_length = Length
      #bus_length = self.cell.bbox().height()*dbu +pitch*2
      constant = math.ceil(Bragg_length/(s1+s2))
      if Bragg_length % (s1+s2) != 0:
        Bragg_length = constant*(s1+s2)   
        
      for ii in range(0,int(row)): 
        xo = [(ii*(w+gap)),(w+ii*(w+gap)),(w+ii*(w+gap)),(ii*(w+gap))]
        yo = [0,0,s1,s1]
        dpts=[pya.DPoint(xo[i], yo[i]) for i in range(len(xo))]
        dpolygon = DPolygon(dpts)
        element = Polygon.from_dpoly(dpolygon*(1.0/dbu))
        shapes(LayerSiN).insert(element)   
        
              
        # draw the SWG waveguide from the center one to +/- positions    
        for i in range(0,int(math.ceil((constant)/2))):
          yu = [yo[j]+i*pitch for j in range(len(yo))]
          yd = [yo[j]-i*pitch for j in range(len(yo))]
            
          dpts=[pya.DPoint(xo[i], yu[i]) for i in range(len(xo))]
          dpolygon = DPolygon(dpts)
          element = Polygon.from_dpoly(dpolygon*(1.0/dbu))
          shapes(LayerSiN).insert(element)
          dpts=[pya.DPoint(xo[i], yd[i]) for i in range(len(xo))]
          dpolygon = DPolygon(dpts)
          element = Polygon.from_dpoly(dpolygon*(1.0/dbu))
          shapes(LayerSiN).insert(element)         
          

        # draw the tapers from waveguide to SWG
        if taperL != 0:
          xtu = [((ii*(w+gap))+(w-0.06)/2),((ii*(w+gap))+(w-0.06)/2+0.06),((ii*(w+gap))+w+gap/2),((ii*(w+gap))-gap/2)]
          ytu = [(yu[3]-taperL),(yu[3]-taperL),(yu[3]),(yu[3])]
          dpts=[pya.DPoint(xtu[i], ytu[i]) for i in range(len(xtu))]
          dpolygon = DPolygon(dpts)
          element = Polygon.from_dpoly(dpolygon*(1.0/dbu))
          shapes(LayerSiN).insert(element)
          
          ytd = [(yd[1]+taperL),(yd[1]+taperL),(yd[1]),(yd[1])]
          dpts=[pya.DPoint(xtu[i], ytd[i]) for i in range(len(xtu))]
          dpolygon = DPolygon(dpts)
          element = Polygon.from_dpoly(dpolygon*(1.0/dbu))
          shapes(LayerSiN).insert(element)
          
          # taper connections from multi-tapers to std WG
          xTu = [(-gap/2),(-gap/2+row*(gap+w)),((row*(gap+w)-gap)/2+0.25),((row*(gap+w)-gap)/2 - 0.25)]
          yTu = [(yu[3]),(yu[3]),(yu[3]+taperL),(yu[3]+taperL)]
          dpts=[pya.DPoint(xTu[i], yTu[i]) for i in range(len(xTu))]
          dpolygon = DPolygon(dpts)
          element = Polygon.from_dpoly(dpolygon*(1.0/dbu))
          shapes(LayerSiN).insert(element)
           
          yTd = [(yd[1]),(yd[1]),(yd[1]-taperL),(yd[1]-taperL)]
          dpts=[pya.DPoint(xTu[i], yTd[i]) for i in range(len(xTu))]
          dpolygon = DPolygon(dpts)
          element = Polygon.from_dpoly(dpolygon*(1.0/dbu))
          shapes(LayerSiN).insert(element)    
          # f is the factor to define the Dev starting point       
          f = 1 
             
             
        if taperL == 0:
          yTd = yd
          yTu = yu
          xTu = [0,0,0,0]
          f = 0
          
      # draw the corrugations of the Bragg gratings
      constant_c = math.ceil((Bragg_length-taperL)/(s1+s2))
          
      for iii in range(0, int(math.ceil(constant_c/4))):
          yuc = [yo[j]+2*iii*pitch for j in range(len(yo))]
          ydc = [yo[j]-2*iii*pitch for j in range(len(yo))]
          xlc = [-wc-s2, -s2, -s2,-wc-s2]
          xrc = [row*pitch, row*pitch+wc, row*pitch+wc, row*pitch]
            
          dpts=[pya.DPoint(xlc[i], yuc[i]) for i in range(len(xo))]
          dpolygon = DPolygon(dpts)
          element = Polygon.from_dpoly(dpolygon*(1.0/dbu))
          shapes(LayerSiN).insert(element)    
              
          dpts=[pya.DPoint(xrc[i], yuc[i]) for i in range(len(xo))]
          dpolygon = DPolygon(dpts)
          element = Polygon.from_dpoly(dpolygon*(1.0/dbu))
          shapes(LayerSiN).insert(element)   
            
      if phaseshifted is False:
        for iii in range(0, int(math.ceil(constant_c/4))):
          #yuc = [yo[j]+2*iii*pitch for j in range(len(yo))]
          ydc = [yo[j]-2*iii*pitch for j in range(len(yo))]
          #xlc = [-wc-s2, -s2, -s2,-wc-s2]
          #xrc = [row*pitch, row*pitch+wc, row*pitch+wc, row*pitch]
            
          dpts=[pya.DPoint(xlc[i], ydc[i]) for i in range(len(xo))]
          dpolygon = DPolygon(dpts)
          element = Polygon.from_dpoly(dpolygon*(1.0/dbu))
          shapes(LayerSiN).insert(element)    
              
          dpts=[pya.DPoint(xrc[i], ydc[i]) for i in range(len(xo))]
          dpolygon = DPolygon(dpts)
          element = Polygon.from_dpoly(dpolygon*(1.0/dbu))
          shapes(LayerSiN).insert(element)
          
      if phaseshifted is True:
        for iii in range(0, int(math.ceil(constant_c/4))):
          #yuc = [yo[j]+2*iii*pitch for j in range(len(yo))]
          yddc = [yo[j]-pitch-2*iii*pitch for j in range(len(yo))]
            
          dpts=[pya.DPoint(xlc[i], yddc[i]) for i in range(len(xo))]
          dpolygon = DPolygon(dpts)
          element = Polygon.from_dpoly(dpolygon*(1.0/dbu))
          shapes(LayerSiN).insert(element)    
              
          dpts=[pya.DPoint(xrc[i], yddc[i]) for i in range(len(xo))]
          dpolygon = DPolygon(dpts)
          element = Polygon.from_dpoly(dpolygon*(1.0/dbu))
          shapes(LayerSiN).insert(element)   
                   

      #DEV BOX
      dev = Box(f*(-gap-wc-(2)*pitch)/dbu, yTd[-1]/dbu, (wc+(row+2)*pitch)/dbu, yTu[-1]/dbu)
      shapes(LayerDevRecN).insert(dev)
      dev_width = self.cell.bbox().width()/2
      
      dev_up = yu[-1]/dbu +taperL/dbu 
      dev_down = yd[-4]/dbu - taperL/dbu
      
      
      # Create the pins on the waveguides, as short paths:
      from SiEPIC._globals import PIN_LENGTH as pin_length
  
      w = to_itype(self.w,dbu)
      gap = to_itype(self.gap,dbu)

      #Pin1
      t = Trans(Trans.R0, xTu[-1]/dbu+250,dev_up)
      pin = Path([Point(0, -pin_length/2), Point(0, pin_length/2)], 500) #500 is width of std WG
      pin_t = pin.transformed(t)
      shapes(LayerPinRecN).insert(pin_t)
      t = Trans(Trans.R270, xTu[-1]/dbu+250,dev_up)
      text = Text ("pin1", t)
      text.halign=0
      shape = shapes(LayerPinRecN).insert(text)
      shape.text_size = 0.4/dbu
 
      #Pin2
      t = Trans(Trans.R0, xTu[-1]/dbu+250,dev_down)
      pin = Path([Point(0, pin_length/2), Point(0, -pin_length/2)], 500)
      pin_t = pin.transformed(t)
      shapes(LayerPinRecN).insert(pin_t)
      t = Trans(Trans.R90, xTu[-1]/dbu+250,dev_down)
      text = Text ("pin2", t)
      text.halign=0
      shape = shapes(LayerPinRecN).insert(text)
      shape.text_size = 0.4/dbu
      
      # Ref
      t = Trans(Trans.R270, xTu[-1]/dbu+250+0.5/dbu,dev_up)
      text = Text ("Ref: E. Luan, doi.org/10.1364/BOE.10.004825", t)
      text.halign=0
      shape = shapes(LayerPinRecN).insert(text)
      shape.text_size = 0.4/dbu




"""
This file is part of the SiEPIC_EBeam_PDK
by Lukas Chrostowski, et al., (c) 2023

This file implements a library called "EBeam-SiN", for silicon nitride


Version history:

Lukas Chrostowski, 2023/01/24
 - Creation of a new library for Silicon Nitride devices
 - directional coupler PCell for 895 nm
 - Added SiN waveguide to WAVEGUIDES.XML

Lukas Chrostowski, 2023/11
 - components for 1550 and 1310
 - waveguides for 1550 and 1310
 - compound waveguide
 - compatibility with PyPI usage of KLayout

"""
verbose=False

if verbose:
    print('siepic_ebeam_library_SiN')

import pya
from pya import *
from SiEPIC.utils import get_technology_by_name
from SiEPIC.utils.layout import make_pin
import os
import pathlib
import sys

dir_path = os.path.dirname(os.path.realpath(__file__))
if dir_path not in sys.path:
    sys.path.append(dir_path)

files = [f for f in os.listdir(os.path.join(os.path.dirname(
    os.path.realpath(__file__)),'pcells_SiN')) if '.py' in pathlib.Path(f).suffixes  and '__init__' not in f]
import pcells_SiN ### folder name ###
import importlib
importlib.invalidate_caches()
pcells_=[]
for f in files:
    module = 'pcells_SiN.%s' % f.replace('.py','')  ### folder name ###
    if verbose:
        print(' - found module: %s' % module)
    m = importlib.import_module(module) 
    if verbose:
        print(m)
    pcells_.append(importlib.reload(m))


class siepic_ebeam_library_hubbard(Library):
  """
  The library where we will put the PCells and GDS into 
  """

  def __init__(self):
  
    
    tech_name = 'EBeam'
    library = tech_name +'-SiN'
    self.technology=tech_name

    if verbose:
        print("Initializing '%s' Library." % library)

    # Set the description
    self.description = "v0.4.4, Silicon Nitride"

    # Save the path, used for loading WAVEGUIDES.XML
    import os
    self.path = os.path.dirname(os.path.realpath(__file__))
                
    # Import all the GDS files from the tech folder
    import os, fnmatch
    dir_path = os.path.normpath(os.path.join(os.path.dirname(os.path.realpath(__file__)), "../gds/EBeam_SiN"))
    if verbose:
        print('  library path: %s' % dir_path)
    search_str = '*.[Oo][Aa][Ss]' # OAS
    for root, dirnames, filenames in os.walk(dir_path, followlinks=True):
        for filename in fnmatch.filter(filenames, search_str):
            file1=os.path.join(root, filename)
            if verbose:
                print(" - reading %s" % file1 )
            self.layout().read(file1)
    search_str = '*.[Gg][Dd][Ss]' # GDS
    for root, dirnames, filenames in os.walk(dir_path, followlinks=True):
        for filename in fnmatch.filter(filenames, search_str):
            file1=os.path.join(root, filename)
            if verbose:
                print(" - reading %s" % file1 )
            self.layout().read(file1)
        
    # Create the PCell declarations
    for m in pcells_:
        mm = m.__name__.replace('pcells_SiN.','')
        mm2 = m.__name__+'.'+mm+'()'
        if verbose:
            print(' - register_pcell %s, %s' % (mm,mm2))
        self.layout().register_pcell(mm, eval(mm2))
    if verbose:
        print(' done with pcells')
    
    # Register us the library with the technology name
    # If a library with that name already existed, it will be replaced then.
    self.register(library)
 
# Instantiate and register the library
siepic_ebeam_library_hubbard()



from pya import *
import pya

class Waveguide_Straight(pya.PCellDeclarationHelper):
  """
  Input: length, width
  draws a straight waveguide with pins. centred at the instantiation point.
  Usage: instantiate, and use transformations (rotation)
  """

  def __init__(self):

    # Important: initialize the super class
    super(Waveguide_Straight, self).__init__()
    from SiEPIC.utils import get_technology_by_name
    TECHNOLOGY = get_technology_by_name('EBeam')

    # declare the parameters
    self.param("wg_length", self.TypeInt, "Waveguide Length", default = 10000)     
    self.param("wg_width", self.TypeInt, "Waveguide width", default = 500)     
    self.param("layer", self.TypeLayer, "Layer", default = TECHNOLOGY['Si'])
    self.param("pinrec", self.TypeLayer, "PinRec Layer", default = TECHNOLOGY['PinRec'])
    self.param("devrec", self.TypeLayer, "DevRec Layer", default = TECHNOLOGY['DevRec'])
  def display_text_impl(self):
    # Provide a descriptive text for the cell
    return "Waveguide_Straight_%.3f-%.3f" % (self.wg_length/1000, self.wg_width/1000)
  
  def coerce_parameters_impl(self):
    pass

  def can_create_from_shape(self, layout, shape, layer):
    return False
    
  def produce_impl(self):
    ly = self.layout
    LayerSiN = ly.layer(self.layer)
    LayerPinRecN = ly.layer(self.pinrec)
    LayerDevRecN = ly.layer(self.devrec)

    # fetch the parameters
    dbu = self.layout.dbu
    ly = self.layout
    shapes = self.cell.shapes

    LayerSi = self.layer
    LayerSiN = ly.layer(LayerSi)
    #LayerPinRecN = ly.layer(self.pinrec)
    LayerDevRecN = ly.layer(self.devrec)

#    print("Waveguide_Straight:")
    w = self.wg_width
    length = self.wg_length
    points = [[-length/2,0], [length/2, 0]]
    path = Path([Point(-length/2,0),Point(length/2,0)],w)
#    print(path)

    shapes(LayerSiN).insert(path.simple_polygon())
    
    from SiEPIC._globals import PIN_LENGTH
    
    # Pins on the bus waveguide side:
    pin_length = PIN_LENGTH
    if length < pin_length+1:
      pin_length = int(length/3)
      pin_length = math.ceil(pin_length / 2.) * 2
    if pin_length == 0:
      pin_length = 2

    t = Trans(Trans.R0,-length/2,0)
    pin = Path([Point(pin_length/2, 0), Point(-pin_length/2, 0)], w)
    pin_t = pin.transformed(t)
    shapes(LayerPinRecN).insert(pin_t)
    text = Text ("pin1", t)
    shape = shapes(LayerPinRecN).insert(text)
    shape.text_size = 0.4/dbu

    t = Trans(Trans.R0,length/2,0)
    pin = Path([Point(-pin_length/2, 0), Point(pin_length/2, 0)], w)
    pin_t = pin.transformed(t)
    shapes(LayerPinRecN).insert(pin_t)
    text = Text ("pin2", t)
    shape = shapes(LayerPinRecN).insert(text)
    shape.text_size = 0.4/dbu
    shape.text_halign = 2

    # Compact model information
    t = Trans(Trans.R0,0, 0)
    text = Text ('Lumerical_INTERCONNECT_library=Design kits/EBeam', t)
    shape = shapes(LayerDevRecN).insert(text)
    shape.text_size = 0.1/dbu
    t = Trans(Trans.R0,length/10, 0)
    text = Text ('Lumerical_INTERCONNECT_component=ebeam_wg_integral_1550', t)
    shape = shapes(LayerDevRecN).insert(text)
    shape.text_size = 0.1/dbu
    t = Trans(Trans.R0,length/9, 0)
    text = Text ('Spice_param:wg_width=%.3fu wg_length=%.3fu'%(self.wg_width*dbu, self.wg_length*dbu), t)
    shape = shapes(LayerDevRecN).insert(text)
    shape.text_size = 0.1/dbu

    # Create the device recognition layer -- make it 1 * wg_width away from the waveguides.
    path = Path([Point(-length/2,0),Point(length/2,0)],w*3)
    shapes(LayerDevRecN).insert(path.simple_polygon())
    


import pya
from pya import *
from SiEPIC.utils import get_technology, get_technology_by_name
from SiEPIC.utils import arc, arc_wg, arc_to_waveguide, points_per_circle#,layout
import math

class phc_H0c_oxide(pya.PCellDeclarationHelper):
  """
  Input: length, width
  """
  import numpy
  
  def __init__(self):

    # Important: initialize the super class
    super(phc_H0c_oxide, self).__init__()

    self.param("a", self.TypeDouble, "lattice constant (microns)", default = 0.690)     
    self.param("n", self.TypeInt, "Number of holes in x and y direction", default = 30)     
    self.param("r", self.TypeDouble, "hole radius (microns)", default = 0.125)
    self.param("wg_dis", self.TypeInt, "Waveguide distance (number of holes)", default = 3)
    self.param("n_bus", self.TypeInt, "Bus number, 1 or 2 ", default = 2)
    self.param("n_vertices", self.TypeInt, "Vertices of a hole", default = 20)                                
    self.param("S1x", self.TypeDouble, "S1x shift", default = 0.28)     
    self.param("S2x", self.TypeDouble, "S2x shift", default = 0.193)     
    self.param("S3x", self.TypeDouble, "S3x shift", default = 0.194)     
    self.param("S4x", self.TypeDouble, "S4x shift", default = 0.162)
    self.param("S5x", self.TypeDouble, "S5x shift", default = 0.113)
    self.param("S1y", self.TypeDouble, "S1y shift", default = -0.016)
    self.param("S2y", self.TypeDouble, "S2y shift", default = 0.134)
    TECHNOLOGY = get_technology_by_name('EBeam')
    self.param("layer", self.TypeLayer, "Layer", default = TECHNOLOGY['Si'])
    self.param("pinrec", self.TypeLayer, "PinRec Layer", default = TECHNOLOGY['PinRec'])
    self.param("devrec", self.TypeLayer, "DevRec Layer", default = TECHNOLOGY['DevRec'])
    self.param("textl", self.TypeLayer, "Text Layer", default = TECHNOLOGY['Text'])

  def display_text_impl(self):
    # Provide a descriptive text for the cell
    return "phc_H0c_oxide_a%s-r%.3f-wg_dis%.3f-n%.3f" % \
    (self.a, self.r, self.wg_dis, self.n)
  
  def coerce_parameters_impl(self):
    pass

  def can_create_from_shape(self, layout, shape, layer):
    return False
    
  def produce_impl(self):
  
    # fetch the parameters
    dbu = self.layout.dbu
    ly = self.layout

    LayerSi = self.layer
    LayerSiN = ly.layer(self.layer)
    LayerPinRecN = ly.layer(self.pinrec)
    LayerDevRecN = ly.layer(self.devrec)
    LayerTextN = ly.layer(self.textl)

    # Fetch all the parameters:
    a = self.a/dbu
    r = self.r/dbu
    wg_dis = self.wg_dis+1
    n_vertices = self.n_vertices
    n_bus = self.n_bus
    n = int(math.ceil(self.n/2))
    Sx = [self.S1x,self.S2x,self.S3x,self.S4x,self.S5x]
    Sy = [self.S1y,0,self.S2y]
    if n_bus == 1:
      Sx = [0,0,0,0,0]
      Sy = [0,0,0]
          
    if wg_dis%2 == 0:
      length_slab_x = (2*n-1)*a
    else:
      length_slab_x = 2*n*a
    
    length_slab_y = 2*(wg_dis+15)*a*math.sqrt(3)/2
    
    n_x = n
    n_y = wg_dis+10

    # Define Si slab and hole region for future subtraction
    Si_slab = pya.Region()
    Si_slab.insert(pya.Box(-length_slab_x/2, -length_slab_y/2, length_slab_x/2, length_slab_y/2))
    hole = pya.Region()
    hole_r = r

    # function to generate points to create a circle
    def circle(x,y,r):
      npts = n_vertices
      theta = 2 * math.pi / npts # increment, in radians
      pts = []
      for i in range(0, npts):
        pts.append(Point.from_dpoint(pya.DPoint((x+r*math.cos(i*theta))/1, (y+r*math.sin(i*theta))/1)))
      return pts
     
    # raster through all holes with shifts and waveguide 
    
    hole_cell = circle(0,0,hole_r)
    hole_poly = pya.Polygon(hole_cell)  

    for j in range(-n_y,n_y+1):
      if j%2 == 0 and j != wg_dis:
        for i in range(-n_x,n_x+1):
          if j == -wg_dis and i > 3 and n_bus == 2:
            None 
          elif j == 0 and i in (1,-1,2,-2,3,-3,4,-4,5,-5):
            hole_x = abs(i)/i*(abs(i)-0.5+Sx[abs(i)-1])*a
            hole_y = 0 
            hole_trans = pya.Trans(Trans.R0, hole_x,hole_y)
            hole_t = hole_poly.transformed(hole_trans)
            hole.insert(hole_t)           
          elif i!=0:
            hole_x = abs(i)/i*(abs(i)-0.5)*a
            hole_y = j*a*math.sqrt(3)/2
            hole_trans = pya.Trans(Trans.R0, hole_x,hole_y)
            hole_t = hole_poly.transformed(hole_trans)
            hole.insert(hole_t) 
      elif j%2 == 1 and j != wg_dis:
        for i in range(-n_x,n_x+1):
          if j == -wg_dis and i > 3 and n_bus == 2:
            None
          elif i == 0 and j in (1,-1,3,-3):
            hole_x = 0
            hole_y = j*a*(math.sqrt(3)/2)+abs(j)/j*a*Sy[abs(j)-1]
            hole_trans = pya.Trans(Trans.R0, hole_x,hole_y)
            hole_t = hole_poly.transformed(hole_trans)
            hole.insert(hole_t) 
          else:  
            hole_x = i*a
            hole_y = j*a*math.sqrt(3)/2
            hole_trans = pya.Trans(Trans.R0, hole_x,hole_y)
            hole_t = hole_poly.transformed(hole_trans)
            hole.insert(hole_t) 
            
    phc = Si_slab - hole
    self.cell.shapes(LayerSiN).insert(phc)
    
    # Pins on the waveguide:    
    pin_length = 200
    pin_w = a
    wg_pos = a*math.sqrt(3)/2*wg_dis
    
    t = pya.Trans(Trans.R0, -length_slab_x/2,wg_pos)
    pin = pya.Path([pya.Point(pin_length/2, 0), pya.Point(-pin_length/2, 0)], pin_w)
    pin_t = pin.transformed(t)
    self.cell.shapes(LayerPinRecN).insert(pin_t)
    text = pya.Text ("pin1", t)
    shape = self.cell.shapes(LayerPinRecN).insert(text)
    shape.text_size = 0.4/dbu

    t = pya.Trans(Trans.R0, length_slab_x/2,wg_pos)
    pin = pya.Path([pya.Point(-pin_length/2, 0), pya.Point(pin_length/2, 0)], pin_w)
    pin_t = pin.transformed(t)
    self.cell.shapes(LayerPinRecN).insert(pin_t)
    text = pya.Text ("pin2", t)
    shape = self.cell.shapes(LayerPinRecN).insert(text)
    shape.text_size = 0.4/dbu
    
    #pin for drop waveguide
    if n_bus == 2:
      t = pya.Trans(Trans.R0, length_slab_x/2,-wg_pos)
      pin_t = pin.transformed(t)
      self.cell.shapes(LayerPinRecN).insert(pin_t)
      text = pya.Text ("pin3", t)
      shape = self.cell.shapes(LayerPinRecN).insert(text)
      shape.text_size = 0.4/dbu

    # Create the device recognition layer -- make it 1 * wg_width away from the waveguides.
    points = [[-length_slab_x/2,0], [length_slab_x/2, 0]]
    points = [Point(each[0], each[1]) for each in points]
    path = Path(points, length_slab_y)   
    self.cell.shapes(LayerDevRecN).insert(path.simple_polygon())
   


"""
This file is part of the SiEPIC-EBeam-PDK

Version history:

Lukas Chrostowski 2023/02/19
 - PCell takes an input of the waveguide type from a dropdown list
    Loaded from Waveguides.xml
 - Metal heater
"""

from pya import *
import pya

class wg_heater(pya.PCellDeclarationHelper):

  def __init__(self):
    # Important: initialize the super class
    super(wg_heater, self).__init__()
    
    from SiEPIC.utils import get_technology_by_name, load_Waveguides_by_Tech

    '''
    # https://github.com/KLayout/klayout/issues/879
    tech = self.layout.library().technology
    if not tech:
       tech = 'EBeam'
    self.technology_name = tech
    '''
    self.technology_name = 'EBeam'
    TECHNOLOGY = get_technology_by_name(self.technology_name)
            
    # Load all waveguides
    self.waveguide_types = load_Waveguides_by_Tech(self.technology_name)   
        
    # declare the parameters

    default_wg = 'SiN Strip TE 895 nm, w=450 nm'
    p = self.param("waveguide_type", self.TypeList, "Waveguide Type", default = default_wg)
#    p = self.param("waveguide_type", self.TypeList, "Waveguide Type", default = self.waveguide_types[0]['name'])

    for wa in self.waveguide_types:
        p.add_choice(wa['name'],wa['name'])
    self.param("length", self.TypeDouble, "Length", default = 100)
    self.param("path", self.TypeShape, "Path", default = DPath([DPoint(0,0), DPoint(100,0)], 0.5), hidden=True)

    self.param("mh_width", self.TypeInt, "Heater width (microns)", default=4)
    self.param("mhlayer", self.TypeLayer, "Heater Layer", default=TECHNOLOGY['M1_heater'])
    self.param("mllayer", self.TypeLayer, "Metal contact layer", default=TECHNOLOGY['M2_router'])
    self.param("metal_width", self.TypeDouble, "Metal contact width", default=10)

    self.param("pinrec", self.TypeLayer, "PinRec Layer", default = TECHNOLOGY['PinRec'])
    self.param("devrec", self.TypeLayer, "DevRec Layer", default = TECHNOLOGY['DevRec'])
    self.param("pinrecm", self.TypeLayer, "PinRecM Layer (metal)", default=TECHNOLOGY['PinRecM'])
    
    self.cellName="wg_heater"

  def display_text_impl(self):
    # Provide a descriptive text for the cell
    return "%s_%s" % (self.cellName, self.length)
  
  def coerce_parameters_impl(self):
    self.path = DPath([DPoint(0,0), DPoint(self.length,0)], 0.5)
    # print(self.waveguide_type)
    
  def can_create_from_shape_impl(self):
    return self.shape.is_path()

  def transformation_from_shape_impl(self):
    return Trans(Trans.R0,0,0)

  def parameters_from_shape_impl(self):
    self.path = self.shape.path
        
  def produce_impl(self):


    # https://github.com/KLayout/klayout/issues/879
    # tech = self.layout.library().technology
        
    # Make sure the technology name is associated with the layout
    #  PCells don't seem to know to whom they belong!
    if self.layout.technology_name == '':
        self.layout.technology_name = self.technology_name

    from SiEPIC.extend import to_itype
    from SiEPIC.utils.layout import make_pin
    dbu = self.layout.dbu
        
    # Draw the waveguide geometry, new function in SiEPIC-Tools v0.3.90
    from SiEPIC.utils.layout import layout_waveguide4
    self.waveguide_length = layout_waveguide4(self.cell, self.path, self.waveguide_type, debug=False)

    # fetch design layers
    LayermlN = self.layout.layer(self.mllayer)
    LayermhN = self.layout.layer(self.mhlayer)
    LayerPinRecN = self.layout.layer(self.pinrec)
    LayerDevRecN = self.layout.layer(self.devrec)
    LayerPinRecM = self.layout.layer(self.pinrecm)

    # draw metal heater
    path = DPath([DPoint(0,0), DPoint(self.length,0)], self.mh_width)
    self.cell.shapes(LayermhN).insert(path.simple_polygon())

    # metal contact
    self.cell.shapes(LayermlN).insert(pya.DPath(
        [pya.DPoint(self.metal_width / 2, -self.mh_width/2), 
         pya.DPoint(self.metal_width / 2, self.mh_width/2)], self.metal_width))
    self.cell.shapes(LayermlN).insert(pya.DPath(
        [pya.DPoint(self.length-self.metal_width / 2, -self.mh_width/2), 
         pya.DPoint(self.length-self.metal_width / 2, self.mh_width/2)], self.metal_width))
    
    # metal pins
    mh_width = to_itype(self.mh_width,dbu)
    ml_width = to_itype(self.metal_width,dbu)
    make_pin(self.cell, "elec1", [self.metal_width / 2, self.mh_width/2], ml_width, LayerPinRecM, 90)
    make_pin(self.cell, "elec2", [self.length-self.metal_width / 2, self.mh_width/2], ml_width, LayerPinRecM, 90)

    # print("EBeam.%s: length %.3f um, complete" % (self.cellName, self.waveguide_length))



import pya
from pya import *
from SiEPIC.utils import get_technology_by_name

class ebeam_rib_phase_shifter_folded(pya.PCellDeclarationHelper):

  # N-doped heater phase shifter in a rib waveguide
  # design by Guowu Zhang <guowu.zhang@mail.mcgill.ca>
  # and Lukas Chrostowski
  # based on Ref: R. Priti, "Efficiency Improvement of an O-band SOI-MZI Thermo-optic Matrix Switch"
  # and folding based on
  # Ref: Z. Lu, "Michelson Interferometer Thermo-optic Switch on SOI with a 50 uW Power Consumption", IEEE PTL, 2015

  def __init__(self):

    # Important: initialize the super class
    super(ebeam_rib_phase_shifter_folded, self).__init__()
    TECHNOLOGY = get_technology_by_name('EBeam')

    # declare the parameters
    self.param("silayer", self.TypeLayer, "Si Layer", default = TECHNOLOGY['Si'])
    self.param("io_wg_type", self.TypeBoolean, "I/O waveguide type, 0 - strip; 1 - rib", default = 0)
    self.param("fold_number", self.TypeInt, "Number of folds (3, 5, 7, ...)", default = 5)   
    self.param("input_rib_width", self.TypeDouble, "Input Rib Width (um)", default = 0.35)
    self.param("input_slab_width", self.TypeDouble, "Input Slab Width (um)", default = 2)
    self.param("in_taper_length", self.TypeDouble, "Input Taper Length (um)", default = 15)    
    self.param("length", self.TypeDouble, "Phase Shifter Length (um)", default = 150)
    self.param("width", self.TypeDouble, "Waveguide Width (um)", default = 0.5)
#    self.param("ps_slab_width", self.TypeDouble, "Phase Shifter Slab Width (um)", default = 18)
    self.param("npp_width", self.TypeDouble, "Npp Width (um)", default = 2)
    self.param("npp_distance", self.TypeDouble, "Npp Edge to Si Edge (um)", default = 2)
    self.param("segments", self.TypeInt, "Phase Shifter Segments", default = 5)  
    self.param("overlay", self.TypeDouble, "Overlay accuracy (optical litho) (um)", default = 1)
    self.param("overlay_ebl", self.TypeDouble, "Overlay accuracy (EBL) (um)", default = 0.05)
#    self.param("vc_dw", self.TypeDouble, "VC Offset", default = 5)
    self.param("vc_w", self.TypeDouble, "Contact via width", default = 5)
#    self.param("m_dw", self.TypeDouble, "Metal Offset", default = 4)
    self.param("m_w", self.TypeDouble, "Metal Width", default = 10)
    self.param("slayer", self.TypeLayer, "Slab Layer", default = TECHNOLOGY['Si - 90 nm rib'])
#    self.param("nlayer", self.TypeLayer, "N Layer", default = TECHNOLOGY['Si N'])
    self.param("npplayer", self.TypeLayer, "N++ Layer", default = TECHNOLOGY['Si N++'])
    self.param("vclayer", self.TypeLayer, "VC Layer", default = TECHNOLOGY['VC'])
    self.param("mlayer", self.TypeLayer, "Metal Layer", default = TECHNOLOGY['M2_router'])
    self.param("pinrec", self.TypeLayer, "PinRec Layer", default = TECHNOLOGY['PinRec'])
    self.param("devrec", self.TypeLayer, "DevRec Layer", default = TECHNOLOGY['DevRec'])
    self.param("textl", self.TypeLayer, "Text Layer", default = LayerInfo(10, 0))

  def display_text_impl(self):
    # Provide a descriptive text for the cell
    return "ebeam_rib_phase_shifter_folded(L=" + ('%.3f' % self.length) + ")"

  def produce(self, layout, layers, parameters, cell):
    
    # coerce parameters (make consistent)
    
    self._layers = layers
    self.cell = cell
    self._param_values = parameters
    self.layout = layout

    # cell: layout cell to place the layout
    # LayerSiN: which layer to use
    # r: radius
    # w: waveguide width
    # length units in dbu

    from math import pi, cos, sin
    from SiEPIC.utils import arc_wg, arc_wg_xy
    from SiEPIC._globals import PIN_LENGTH
    from SiEPIC.utils.layout import layout_waveguide2
    from SiEPIC.extend import to_itype

    # fetch the parameters
    TECHNOLOGY = get_technology_by_name('EBeam')
    dbu = self.layout.dbu
    ly = self.layout
        
    LayerSiN = ly.layer(self.silayer)
    LayerSlab = ly.layer(self.slayer)
#    LayerNN = ly.layer(self.nlayer)
    LayerNPPN = ly.layer(self.npplayer)
    LayerVCN = ly.layer(self.vclayer)
    LayerMN = ly.layer(self.mlayer)
    LayerPinRecN = ly.layer(self.pinrec)
    LayerDevRecN = ly.layer(self.devrec)
    TextLayerN = ly.layer(self.textl)
     
    ## define parameters for phase shifter
    overlay = to_itype(self.overlay,dbu)
    overlay_ebl = to_itype(self.overlay_ebl,dbu)
    
    ## define waveguide related parameters
    radius_um, adiab, bezier = 5, 1, 0.2
    w = to_itype(self.width,dbu)   #waveguide width, default 0.5 um
    l = to_itype(self.length,dbu)  #pahse shifter length, default 150 um
    npp_d = to_itype(self.npp_distance, dbu)    #npp to center waveguide distance
    sw = (npp_d+w/2) * 2    #twice of npp_d that is defined above
    in_rib = to_itype(self.input_rib_width,dbu)   #the slab width of the input/output waveguide to the phase shifter
    in_taper = to_itype(self.in_taper_length,dbu) #the input waveguide to phase shifter dimension transfer taper length, default 15
    edge_slab_width = to_itype(4,dbu) #width of the 90 nm slab that is outside of all the folding waveguide, default 2 * 2um
    npp_width = to_itype(self.npp_width,dbu)   #npp width, default 5 um
    
    ## define waveguide folding parameters
    folding_n = int(round((self.fold_number - 1)/2)) # folding waveguide number on one side
    folding_w1 = w+50   # one of the folding waveguide width
    folding_w2 = w-50 # the other one of the folding waveguide width
    folding_gap = to_itype(2,dbu)  # the gap between two folding waveguide
    routing_step = to_itype(radius_um,dbu) # define how large the U shape should be, default 20
    
    ## define node related parameters
    contact_size = to_itype(self.vc_w,dbu)  # square NPP size, for VC purpose
    pin_text_size = to_itype(0.4,dbu)  # pin text size
    
    metal_routing_distance_to_node = to_itype(30,dbu)  # *** calculate from PCell parameters ***
    metal_routing_width = to_itype(self.m_w, dbu) # width of metal
    vc_to_npp_exclusion = to_itype(self.overlay,dbu) # VC boundary to NPP boundary distance

    ps_sl_w = sw+2*contact_size+folding_n*2*(w+folding_gap+overlay_ebl)  # overall pahse shifter 90nm slab width
    
    # measure the total length of the waveguides
    total_waveguide_length = 0

    ## define strip to slab transfer taper parameters
    N = 100; # Number of points for the input/output slab taper
    order = 3; # input/output slab taper curve 

    pts = [Point(-l/2,0), Point(l/2, 0)]
    total_waveguide_length += layout_waveguide2(TECHNOLOGY, self.layout, self.cell, ['Si'], [w*dbu], [0], pts, radius_um, adiab, bezier)

    wg2 = pya.Box(-l/2,-ps_sl_w/2,l/2,ps_sl_w/2)
    wg3 = pya.Box(-l/2,npp_d+w/2,l/2, npp_d+w/2 + npp_width)
    wg4 = pya.Box(-l/2,-npp_d-w/2,l/2, -npp_d-w/2 - npp_width)


    self.cell.shapes(LayerNPPN).insert(wg3)
    self.cell.shapes(LayerNPPN).insert(wg4)
    
    if self.io_wg_type:
      in_slab = to_itype(self.input_slab_width,dbu)
    else:
      in_slab = in_rib
  
    if folding_n > 0:
      ps_sl_w = contact_size*2 + sw * 2 + (folding_w1 + folding_w2 + folding_gap * 2) * folding_n + edge_slab_width
      wg2 = pya.Box(-l/2,-ps_sl_w/2,l/2,ps_sl_w/2)
      self.cell.shapes(LayerSlab).insert(wg2)

      pts = [Point(-l/2 - in_taper, -w/2), Point( -l/2, -w/2), Point(-l/2, w/2), Point(-l/2 - in_taper, w/2)] # fix waveguide width mismatch for left center input taper
      self.cell.shapes(LayerSiN).insert(Polygon(pts))
  
      # add input slab taper
      pts = [];
      for i in range(0,N + 1):
        pts.append(Point(-l/2 - in_taper + in_taper/N * i, w/2 + ((sw/2 + contact_size - w)/(N**order))*(i**order)))
      for i in range(0,N + 1):
        pts.append(Point(-l/2 - in_taper + in_taper/N * (N - i), - w/2 - ((sw/2 + contact_size - w)/(N**order))*((N-i)**order)))
      self.cell.shapes(LayerSlab).insert(Polygon(pts))
  
      # add output strip taper
      pts = [Point(l/2 + in_taper, -w/2), Point( l/2, -w/2), Point(l/2, w/2), Point(l/2 + in_taper, w/2)]
      #wg2 = pya.Box(-l/2,-4/2/dbu,l/2,4/2/dbu)
      self.cell.shapes(LayerSiN).insert(Polygon(pts))
      total_waveguide_length += in_taper*dbu
  
      # add input slab taper
      pts = [];
      for i in range(0,N + 1):
        pts.append(Point(l/2 + in_taper - in_taper/N * i, (w-overlay_ebl)/2 + ((sw/2 + contact_size - (w-overlay_ebl))/(N**order))*(i**order)))
      for i in range(0,N + 1):
        pts.append(Point(l/2 + in_taper/N * i, - (w-overlay_ebl)/2 - ((sw/2 + contact_size - (w-overlay_ebl))/(N**order))*((N-i)**order)))
      self.cell.shapes(LayerSlab).insert(Polygon(pts))

      for i in range(0, folding_n):
        y_coordinate = contact_size + sw + 2 * vc_to_npp_exclusion # move up/down the position according to vc_to_npp_exclusion
        y_move = i * (folding_w1 + folding_w2 + folding_gap * 2)/2 +  y_coordinate
        if i%2:
          folding_w = folding_w2
        else:
          folding_w = folding_w1
                 
        # waveguide in the heated region 
        pts = [Point(-l/2,y_move), Point(l/2, y_move)]
        total_waveguide_length += layout_waveguide2(TECHNOLOGY, self.layout, self.cell, ['Si'], [folding_w*dbu], [0], pts, radius_um, adiab, bezier)
        pts = [Point(-l/2,- y_move), Point(l/2,- y_move)]
        total_waveguide_length += layout_waveguide2(TECHNOLOGY, self.layout, self.cell, ['Si'], [folding_w*dbu], [0], pts, radius_um, adiab, bezier)


        
        # add upper left input taper
        pts = [Point(-l/2 - in_taper, -w/2 + y_move), Point( -l/2, -folding_w/2 + y_move), Point(-l/2, folding_w/2 + y_move), Point(-l/2 - in_taper, w/2 + y_move)]
        #wg2 = pya.Box(-l/2,-4/2/dbu,l/2,4/2/dbu)
        self.cell.shapes(LayerSiN).insert(Polygon(pts))        
        total_waveguide_length += in_taper*dbu
        
        # slab cubic taper
        pts = [];
        for ii in range(0,N + 1):
          pts.append(Point(-l/2 - in_taper + in_taper/N * ii, (w-overlay_ebl)/2 + ((2/dbu - (w-overlay_ebl))/(N**order))*(ii**order)+y_move))
        for ii in range(0,N + 1):
          pts.append(Point(-l/2 - in_taper + in_taper/N * (N - ii), - (w-overlay_ebl)/2 - ((2/dbu - (w-overlay_ebl))/(N**order))*((N-ii)**order)+y_move))
        self.cell.shapes(LayerSlab).insert(Polygon(pts))      
        
                
        if (i!=folding_n-1):   
          # add below left taper   
          y_move = - y_move
          pts = [Point(-l/2 - in_taper, -w/2 + y_move), Point( -l/2, -folding_w/2 + y_move), Point(-l/2, folding_w/2 + y_move), Point(-l/2 - in_taper, w/2 + y_move)]
          #wg2 = pya.Box(-l/2,-4/2/dbu,l/2,4/2/dbu)
          self.cell.shapes(LayerSiN).insert(Polygon(pts))        
          total_waveguide_length += in_taper*dbu
          
          # slab cubic taper
          pts = [];
          for ii in range(0,N + 1):
            pts.append(Point(-l/2 - in_taper + in_taper/N * ii, (w-overlay_ebl)/2 + ((2/dbu - (w-overlay_ebl))/(N**order))*(ii**order)+
            y_move))
          for ii in range(0,N + 1):
            pts.append(Point(-l/2 - in_taper + in_taper/N * (N - ii), - (w-overlay_ebl)/2 - ((2/dbu - (w-overlay_ebl))/(N**order))*((N-ii)**order)+
            y_move))
          self.cell.shapes(LayerSlab).insert(Polygon(pts))     
        
          # add upper right taper
          y_move = - y_move 
          pts = [Point(l/2 + in_taper, -w/2 + y_move), Point( l/2, -folding_w/2 + y_move), Point(l/2, folding_w/2 + y_move), Point(l/2 + in_taper, w/2 + y_move)]
          #wg2 = pya.Box(-l/2,-4/2/dbu,l/2,4/2/dbu)
          self.cell.shapes(LayerSiN).insert(Polygon(pts))
          total_waveguide_length += in_taper*dbu
      
          # slab cubic taper
          pts = [];
          for ii in range(0,N + 1):
            pts.append(Point(l/2 + in_taper - in_taper/N * ii, (w-overlay_ebl)/2 + ((2/dbu - (w-overlay_ebl))/(N**order))*(ii**order) + y_move))
          for ii in range(0,N + 1):
            pts.append(Point(l/2 + in_taper/N * ii, - (w-overlay_ebl)/2 - ((2/dbu - (w-overlay_ebl))/(N**order))*((N-ii)**order) + y_move))
          self.cell.shapes(LayerSlab).insert(Polygon(pts)) 

        else:
          y_move = -y_move
          # add input strip taper
          pts = [Point(-l/2 - in_taper, -w/2 + y_move), Point( -l/2, -folding_w/2 + y_move), 
          Point(-l/2, folding_w/2 + y_move), Point(-l/2 - in_taper, w/2 + y_move)]
          #wg2 = pya.Box(-l/2,-4/2/dbu,l/2,4/2/dbu)
          self.cell.shapes(LayerSiN).insert(Polygon(pts))
          total_waveguide_length += in_taper*dbu
      
          # add input slab taper
          pts = [];
          for i in range(0,N + 1):
            pts.append(Point(-l/2 - in_taper + in_taper/N * i, in_slab/2 + ((edge_slab_width - in_slab)/(N**order))*(i**order) + y_move))
          for i in range(0,N + 1):
            pts.append(Point(-l/2 - in_taper + in_taper/N * (N - i), - in_slab/2 - ((edge_slab_width - in_slab)/(N**order))*((N-i)**order) + y_move))
          self.cell.shapes(LayerSlab).insert(Polygon(pts))
      
          # add output strip taper
          y_move = - y_move
          pts = [Point(l/2 + in_taper, -w/2 + y_move), Point( l/2, -folding_w/2 + y_move), Point(l/2, folding_w/2 + y_move), Point(l/2 + in_taper, w/2 + y_move)]
          #wg2 = pya.Box(-l/2,-4/2/dbu,l/2,4/2/dbu)
          self.cell.shapes(LayerSiN).insert(Polygon(pts))
          total_waveguide_length += in_taper*dbu
      
          # add input slab taper
          pts = [];
          for i in range(0,N + 1):
            pts.append(Point(l/2 + in_taper - in_taper/N * i, in_slab/2 + ((edge_slab_width - in_slab)/(N**order))*(i**order) + y_move))
          for i in range(0,N + 1):
            pts.append(Point(l/2 + in_taper/N * i, - in_slab/2 - ((edge_slab_width - in_slab)/(N**order))*((N-i)**order) + y_move))
          self.cell.shapes(LayerSlab).insert(Polygon(pts))
      
          # Pin on the left side:
          wg_start = - ( l/2 + in_taper+routing_step*folding_n*3+w)
          p1 = [Point(wg_start + PIN_LENGTH/2,- y_move), Point(wg_start - PIN_LENGTH/2,- y_move)]
          p1c = Point(wg_start,- y_move)
          self.set_p1 = p1c
          self.p1 = p1c
          pin = Path(p1, w)
          self.cell.shapes(LayerPinRecN).insert(pin)
          t = Trans(Trans.R0, wg_start, - y_move)
          text = Text ("pin1", t)
          shape = self.cell.shapes(LayerPinRecN).insert(text)
          shape.text_size = pin_text_size
          # Waveguide to route to port, new in SiEPIC-Tools v0.3.64
          pts = [Point(-l/2 - in_taper, -y_move), 
            Point(wg_start, -y_move)]
          if self.io_wg_type:
              waveguide_length = layout_waveguide2(TECHNOLOGY, self.layout, self.cell, ['Si', 'Si - 90 nm rib'], [w*dbu, in_slab*dbu], [0,0], pts, radius_um, adiab, bezier)
          else:
              waveguide_length = layout_waveguide2(TECHNOLOGY, self.layout, self.cell, ['Si'], [w*dbu], [0], pts, radius_um, adiab, bezier)
      
          # Pin & Waveguide on the right side:
          wg_end = l/2 + in_taper+routing_step*folding_n*3+w
          p2 = [Point(wg_end - PIN_LENGTH/2,y_move), Point(wg_end + PIN_LENGTH/2,y_move)]
          p2c = Point(wg_end, y_move)
          self.set_p2 = p2c
          self.p2 = p2c
          pin = Path(p2, w)
          self.cell.shapes(LayerPinRecN).insert(pin)
          t = Trans(Trans.R0, wg_end, y_move)
          text = Text ("pin2", t)
          shape = self.cell.shapes(LayerPinRecN).insert(text)
          shape.text_size = pin_text_size  
          # Waveguide to route to port, new in SiEPIC-Tools v0.3.64
          pts = [Point(l/2 + in_taper, y_move), 
            Point(wg_end, y_move)]
          if self.io_wg_type:
              total_waveguide_length += layout_waveguide2(TECHNOLOGY, self.layout, self.cell, ['Si', 'Si - 90 nm rib'], [w*dbu, in_slab*dbu], [0,0], pts, radius_um, adiab, bezier)
          else:
              total_waveguide_length += layout_waveguide2(TECHNOLOGY, self.layout, self.cell, ['Si'], [w*dbu], [0], pts, radius_um, adiab, bezier)

        # add below right taper   
        y_move = - y_move 
        pts = [Point(l/2 + in_taper, -w/2 + y_move), Point( l/2, -folding_w/2 + y_move), Point(l/2, folding_w/2 + y_move), Point(l/2 + in_taper, w/2 + y_move)]
        #wg2 = pya.Box(-l/2,-4/2/dbu,l/2,4/2/dbu)
        self.cell.shapes(LayerSiN).insert(Polygon(pts))
        total_waveguide_length += in_taper*dbu
        pts = [];
        for ii in range(0,N + 1):
          pts.append(Point(l/2 + in_taper - in_taper/N * ii, (w-overlay_ebl)/2 + ((2/dbu - (w-overlay_ebl))/(N**order))*(ii**order) + y_move))
        for ii in range(0,N + 1):
          pts.append(Point(l/2 + in_taper/N * ii, - (w-overlay_ebl)/2 - ((2/dbu - (w-overlay_ebl))/(N**order))*((N-ii)**order) + y_move))
        self.cell.shapes(LayerSlab).insert(Polygon(pts))   
        
      # add metal contact and folded waveguides
      for i in range(-folding_n, folding_n):
        y_coordinate = contact_size + sw + 2 * vc_to_npp_exclusion # move up/down the position according to vc_to_npp_exclusion
        if i<0:
          y_origin = (i + 1) * (folding_w1 + folding_w2 + folding_gap * 2)/2 -  y_coordinate
        if i==0:
          y_origin = 0   
        if i>0:
          y_origin = (i - 1) * (folding_w1 + folding_w2 + folding_gap * 2)/2 +  y_coordinate      
        
        if (i+1)<0:
          y_end = (i + 2) * (folding_w1 + folding_w2 + folding_gap * 2)/2 -  y_coordinate
        if (i+1)==0:
          y_end = 0   
        if (i+1)>0:
          y_end = i * (folding_w1 + folding_w2 + folding_gap * 2)/2 +  y_coordinate   
        
        # Calculate the Folded waveguide vertices
        if (i+folding_n)%2:
          # Waveguides on the left side
          if y_origin + routing_step * 2 - y_end < 2*radius_um:
            y_routing = max(routing_step*2,y_end-y_origin) 
          else:
            y_routing = y_end+routing_step*2
          path = Path([Point(-l/2 - in_taper, y_origin), 
            Point(-l/2 - in_taper - routing_step * (1.5*folding_n - 1.5*i + 1.5), y_origin),
            Point(-l/2 - in_taper - routing_step * (1.5*folding_n - 1.5*i + 1.5), y_routing), 
            Point(-l/2 - in_taper - routing_step * (1.5*folding_n - 1.5*i - 0.5), y_routing), 
            Point(-l/2 - in_taper - routing_step * (1.5*folding_n - 1.5*i - 0.5), y_end), 
            Point(-l/2 - in_taper, y_end)],0)
           
        else:  
          # Waveguides on the right side
          if y_origin - y_end + routing_step * 2 < 2*radius_um:
            y_routing = -max(routing_step*2, y_end-y_origin)
          else:
            y_routing = y_origin-routing_step*2
          path = Path([Point(l/2 + in_taper, y_origin), 
            Point(l/2 + in_taper + routing_step * (1.5*folding_n + 1.5*i+1), y_origin), 
            Point(l/2 + in_taper + routing_step * (1.5*folding_n + 1.5*i+1), y_routing), 
            Point(l/2 + in_taper + routing_step * (1.5*folding_n + 1.5*i+3), y_routing), 
            Point(l/2 + in_taper + routing_step * (1.5*folding_n + 1.5*i+3), y_end), 
            Point(l/2 + in_taper, y_end)], 0)

        # Draw the waveguide geometry, new in SiEPIC-Tools v0.3.64
        path.unique_points()
        pts = path.get_points()
        total_waveguide_length += layout_waveguide2(TECHNOLOGY, self.layout, self.cell, ['Si'],  [w*dbu], [0], pts, radius_um, adiab, bezier)
           


    for i in range(0, self.segments + 1):
        
        wg_temp_upper = pya.Box((l - contact_size)/self.segments *  i - l/2 - vc_to_npp_exclusion, sw/2,(l - contact_size)/self.segments *  i - l/2 + contact_size + vc_to_npp_exclusion, sw/2 + contact_size + 2 * vc_to_npp_exclusion)
        #self.cell.shapes(LayerNN).insert(wg_temp_upper)
        self.cell.shapes(LayerNPPN).insert(wg_temp_upper)
#        self.cell.shapes(LayerNN).insert(wg_temp_upper)
        wg_temp_upper = pya.Box((l - contact_size)/self.segments *  i - l/2 - 2 * vc_to_npp_exclusion, sw/2,(l - contact_size)/self.segments *  i - l/2 + contact_size + 2 * vc_to_npp_exclusion, sw/2 + contact_size + 3 * vc_to_npp_exclusion)
        self.cell.shapes(LayerSlab).insert(wg_temp_upper)
        
        vc_temp_upper = pya.Box((l - contact_size)/self.segments *  i - l/2, sw/2 + vc_to_npp_exclusion,(l - contact_size)/self.segments *  i - l/2 + contact_size, sw/2 + contact_size + vc_to_npp_exclusion)
        self.cell.shapes(LayerVCN).insert(vc_temp_upper)
        
        wg_temp_lower = pya.Box((l - contact_size)/self.segments *  i - l/2 - vc_to_npp_exclusion, -sw/2,(l - contact_size)/self.segments *  i - l/2 + contact_size + vc_to_npp_exclusion, -sw/2 - contact_size - 2 * vc_to_npp_exclusion)
        #self.cell.shapes(LayerNN).insert(wg_temp_lower)
        self.cell.shapes(LayerNPPN).insert(wg_temp_lower)
        wg_temp_lower = pya.Box((l - contact_size)/self.segments *  i - l/2 - 2 * vc_to_npp_exclusion, -sw/2,(l - contact_size)/self.segments *  i - l/2 + contact_size + 2 * vc_to_npp_exclusion, -sw/2 - contact_size - 3 * vc_to_npp_exclusion)
        self.cell.shapes(LayerSlab).insert(wg_temp_lower)

#        self.cell.shapes(LayerNN).insert(wg_temp_lower)
        vc_temp_lower = pya.Box((l - contact_size)/self.segments *  i - l/2, -sw/2 - vc_to_npp_exclusion,(l - contact_size)/self.segments *  i - l/2 + contact_size, -sw/2 - contact_size - vc_to_npp_exclusion)
        self.cell.shapes(LayerVCN).insert(vc_temp_lower)

        metal_temp = pya.Box((l - contact_size)/self.segments *  i - l/2 - vc_to_npp_exclusion, -sw/2 - contact_size - metal_routing_distance_to_node * (i%2) - 2 * vc_to_npp_exclusion,(l - contact_size)/self.segments *  i - l/2 + contact_size + vc_to_npp_exclusion, sw/2 + contact_size + metal_routing_distance_to_node * ((i+1)%2) + 2 * vc_to_npp_exclusion)
        self.cell.shapes(LayerMN).insert(metal_temp)
        
            
    metal_routing = Path([Point(-l/2, -sw/2 - metal_routing_distance_to_node - contact_size), 
    Point(l/2, -sw/2 - metal_routing_distance_to_node - contact_size)], metal_routing_width)
    self.cell.shapes(LayerMN).insert(metal_routing)
    metal_routing = Path([Point(-l/2, sw/2 + metal_routing_distance_to_node + contact_size), 
    Point(l/2, sw/2 + metal_routing_distance_to_node + contact_size)], metal_routing_width)
    self.cell.shapes(LayerMN).insert(metal_routing)

    if folding_n > 0:
      dev = Box(wg_start, -sw/2 - metal_routing_distance_to_node - contact_size - metal_routing_width,
      wg_end, sw/2 + metal_routing_distance_to_node + contact_size + metal_routing_width)    
      self.cell.shapes(LayerDevRecN).insert(dev)

    t = Trans(0, False, Point(0,0)) 
    text = Text ( \
      'Waveguide Length=%.3fu' %(total_waveguide_length), t, 3*w, -1)
    shape = self.cell.shapes(LayerDevRecN).insert(text)



import pya
import math
from pya import *
from SiEPIC.utils import get_technology_by_name

def layout_Ring(cell, layer, x, y, r, w, npoints):
    
    # function to produce the layout of a ring resonator
    # cell: layout cell to place the layout
    # layer: which layer to use
    # x, y: location of the origin
    # r: radius
    # w: waveguide width
    # units in microns

    # example usage.  Places the ring layout in the presently selected cell.
    # cell = Application.instance().main_window().current_view().active_cellview().cell
    # layout_Ring(cell, cell.layout().layer(TECHNOLOGY['Si']), 0, 0, 10, 0.5, 400)


    # fetch the database parameters
    dbu = cell.layout().dbu
    
    # compute the circle
    pts = []
    da = math.pi * 2 / npoints
    for i in range(0, npoints+1):
      pts.append(Point.from_dpoint(DPoint((x+(r+w/2)*math.cos(i*da))/dbu, (y+(r+w/2)*math.sin(i*da))/dbu)))
    for i in range(npoints, -1, -1):
      pts.append(Point.from_dpoint(DPoint((x+(r-w/2)*math.cos(i*da))/dbu, (y+(r-w/2)*math.sin(i*da))/dbu)))
    
    # create the shape
    cell.shapes(layer).insert(Polygon(pts))

    # end of layout_Ring


class DoubleBus_Ring(pya.PCellDeclarationHelper):
  """
  The PCell declaration for the double bus ring resonator.
  Consists of a ring with 2 straight waveguides.
  """

  def __init__(self):

    # Important: initialize the super class
    super(DoubleBus_Ring, self).__init__()
    TECHNOLOGY = get_technology_by_name('EBeam')

    # declare the parameters
    self.param("silayer", self.TypeLayer, "Si Layer", default = TECHNOLOGY['Si'])
    self.param("s", self.TypeShape, "", default = DPoint(0, 0))
    self.param("r", self.TypeDouble, "Radius", default = 10)
    self.param("w", self.TypeDouble, "Waveguide Width", default = 0.5)
    self.param("g", self.TypeDouble, "Gap", default = 0.2)
    self.param("textpolygon", self.TypeInt, "Draw text polygon label? 0/1", default = 1)
    self.param("textl", self.TypeLayer, "Text Layer", default = LayerInfo(10, 0))
    self.param("pinrec", self.TypeLayer, "PinRec Layer", default = TECHNOLOGY['PinRec'])
    self.param("devrec", self.TypeLayer, "DevRec Layer", default = TECHNOLOGY['DevRec'])

  def display_text_impl(self):
    # Provide a descriptive text for the cell
    return "DoubleBus_Ring(R=" + ('%.3f' % self.r) + ",g=" + ('%g' % (1000*self.g)) + ")"

  def can_create_from_shape_impl(self):
    return False
    
  def produce_impl(self):
    # This is the main part of the implementation: create the layout

    # fetch the parameters
    dbu = self.layout.dbu
    ly = self.layout
    shapes = self.cell.shapes
    
    LayerSi = self.silayer
    LayerSiN = ly.layer(LayerSi)
    TextLayerN = ly.layer(self.textl)
    LayerPinRecN = ly.layer(self.pinrec)
    LayerDevRecN = ly.layer(self.devrec)

    from SiEPIC.utils import points_per_circle, layout_pgtext
    
    # Create the ring resonator:
    layout_Ring(self.cell, LayerSiN, self.r+self.w/2, self.r+self.g+self.w, self.r, self.w, points_per_circle(self.r))

    w = int(round(self.w/dbu))
    r = int(round(self.r/dbu))
    g = int(round(self.g/dbu))

 #   pcell = ly.create_cell("DirectionalCoupler_HalfRing_Straight", "SiEPIC", { "r": self.r, "w": self.w, "g": self.g, "silayer": LayerSi, "bustype": 0 } )
 #   print ("Cell: pcell: #%s" % pcell.cell_index())
 #   t = Trans(Trans.R0, 0, 0) 
 #   instance = self.cell.insert(CellInstArray(pcell.cell_index(), t))
 #   t = Trans(Trans.R180, 0, 2*r+2*g+2*w) 
 #   instance = self.cell.insert(CellInstArray(pcell.cell_index(), t))


    # Create the two waveguides
    wg1 = Box(0, -w/2, w+2*r, w/2)
    shapes(LayerSiN).insert(wg1)
    y_offset = 2*r + 2*g + 2*w
    wg2 = Box(0, y_offset-w/2, w+2*r, y_offset+w/2)
    shapes(LayerSiN).insert(wg2)

    from SiEPIC._globals import PIN_LENGTH as pin_length
    # Create the pins, as short paths:
    
    pin = Path([Point(pin_length/2, 0), Point(-pin_length/2, 0)], w)
    shapes(LayerPinRecN).insert(pin)
    t = Trans(Trans.R0, 0, 0)
    text = Text ("pin1", t)
    shape = shapes(LayerPinRecN).insert(text)
    shape.text_size = 0.4/dbu

    pin = Path([Point(w+2*r-pin_length/2, 0), Point(w+2*r+pin_length/2, 0)],w )
    shapes(LayerPinRecN).insert(pin)
    t = Trans(Trans.R0, w+2*r, 0)
    text = Text ("pin2", t)
    shape = shapes(LayerPinRecN).insert(text)
    shape.text_size = 0.4/dbu

    pin = Path([Point(pin_length/2, y_offset), Point(-pin_length/2, y_offset)], w)
    shapes(LayerPinRecN).insert(pin)
    t = Trans(Trans.R0, 0, y_offset)
    text = Text ("pin3", t)
    shape = shapes(LayerPinRecN).insert(text)
    shape.text_size = 0.4/dbu

    pin = Path([Point(w+2*r-pin_length/2, y_offset), Point(w+2*r+pin_length/2, y_offset)], w)
    shapes(LayerPinRecN).insert(pin)
    t = Trans(Trans.R0, w+2*r, y_offset)
    text = Text ("pin4", t)
    shape = shapes(LayerPinRecN).insert(text)
    shape.text_size = 0.4/dbu


    # Create the device recognition layer
    dev = Box(0, -w*3, w+2*r, y_offset+w*3 )
    shapes(LayerDevRecN).insert(dev)


    # Add a polygon text description
    if self.textpolygon:
      layout_pgtext(self.cell, self.textl, self.w, self.r+self.w, "%.3f-%g" % ( self.r, self.g), 1)

    # print("Done drawing the layout for - DoubleBus_Ring: %.3f-%g" % ( self.r, self.g) )



import pya
from pya import *
from SiEPIC.utils import get_technology_by_name

class ebeam_irph_mrr(pya.PCellDeclarationHelper):
  """
  The PCell declaration for the IRPH MRR.
  
  Author: Jaspreet Jhoja
  jaspreetj@ece.ubc.ca
  """

  def __init__(self):

    # Important: initialize the super class
    super(ebeam_irph_mrr, self).__init__()
    TECHNOLOGY = get_technology_by_name('EBeam')

    # declare the parameters
    self.param("silayer", self.TypeLayer, "Si Layer", default = TECHNOLOGY['Si'])
    self.param("siriblayer", self.TypeLayer, "Si rib Layer", default = TECHNOLOGY['Si - 90 nm rib'])
    self.param("r", self.TypeDouble, "Radius", default = 15)
    self.param("w", self.TypeDouble, "Waveguide Width", default = 0.5)
    self.param("g", self.TypeDouble, "Gap", default = 0.156)
    self.param("io_wg_type", self.TypeBoolean, "I/O waveguide type, 0 - strip; 1 - rib", default = 0)
    
    self.param("n_w", self.TypeDouble, "N Width", default = 31)
    self.param("n_theta_start", self.TypeDouble, "Start Theta", default = -25)
    self.param("n_theta_stop", self.TypeDouble, "Stop Theta", default = 115)
    
    self.param("npp_si", self.TypeDouble, "N++ to Si edge", default = 2.0)
    self.param("npp_w", self.TypeDouble, "N++ Donut Width", default = 13.5)
    
    self.param("vc_cw", self.TypeDouble, "VC Center radius", default = 5.75)
    self.param("vc_aw", self.TypeDouble, "VC Arc Width", default = 5)

    self.param("m_cw", self.TypeDouble, "Metal Center radius", default = 10)
    self.param("m_aw", self.TypeDouble, "Metal Arc Width", default = 11)
    self.param("overlay_ebl", self.TypeDouble, "Overlay accuracy (EBL) (um)", default = 0.05)
        
    self.param("nlayer", self.TypeLayer, "N Layer", default = TECHNOLOGY['Si N'])
    self.param("npplayer", self.TypeLayer, "N++ Layer", default = TECHNOLOGY['Si N++'])
    self.param("vclayer", self.TypeLayer, "VC Layer", default = TECHNOLOGY['VC'])
    self.param("mlayer", self.TypeLayer, "Metal Layer", default = TECHNOLOGY['M2_router'])
    self.param("pinrec", self.TypeLayer, "PinRec Layer", default = TECHNOLOGY['PinRec'])
    self.param("devrec", self.TypeLayer, "DevRec Layer", default = TECHNOLOGY['DevRec'])
    self.param("textl", self.TypeLayer, "Text Layer", default = LayerInfo(10, 0))

  def display_text_impl(self):
    # Provide a descriptive text for the cell
    return "ebeam_irph_mrr(R=" + ('%.3f' % self.r) + ",g=" + ('%g' % (1000*self.g)) + ")"

  def can_create_from_shape_impl(self):
    return False
    
  def produce_impl(self):
    # This is the main part of the implementation: create the layout

    from math import pi, cos, sin
    from SiEPIC.utils import arc_wg, arc_wg_xy, arc_xy
    from SiEPIC._globals import PIN_LENGTH
    from SiEPIC.extend import to_itype, to_dtype

    # fetch the parameters
    dbu = self.layout.dbu
    ly = self.layout
    shapes = self.cell.shapes
    
    LayerSiN = ly.layer(self.silayer)
    LayerSiRibN = ly.layer(self.siriblayer)
    LayerNN = ly.layer(self.nlayer)
    LayerNPPN = ly.layer(self.npplayer)
    LayerVCN = ly.layer(self.vclayer)
    LayerMN = ly.layer(self.mlayer)
    LayerPinRecN = ly.layer(self.pinrec)
    LayerDevRecN = ly.layer(self.devrec)
    TextLayerN = ly.layer(self.textl)
    
    w = to_itype(self.w,dbu)
    r = to_itype(self.r,dbu)
    g = to_itype(self.g,dbu)
    
    n_w = to_itype(self.n_w,dbu)
    npp_si = to_itype(self.npp_si, dbu)
    npp_w=to_itype(self.npp_w,dbu)
    theta_start = self.n_theta_start
    theta_stop = self.n_theta_stop

    vc_cw = to_itype(self.vc_cw,dbu)
    vc_aw = to_itype(self.vc_aw,dbu)
    overlay_ebl = to_itype(self.overlay_ebl,dbu)
    
    m_cw = to_itype(self.m_cw,dbu)
    m_aw = to_itype(self.m_aw,dbu)
    x = 0
    y = 0
    
    #draw ring
    self.cell.shapes(LayerSiN).insert(arc_wg_xy(x, y, r, w, 0, 360))
        
    #draw bus waveguides  
    #bottom waveguide
    xtop = -2/3*r
    ytop = -1*(r+g+w/2)
    xbottom = 4/3*r
    ybottom = ytop-w
    wg1 = Box(xtop, ytop , xbottom, ybottom)
    shapes(LayerSiN).insert(wg1)
    #left waveguide
    wg1 = Box(ytop, xtop , ybottom, xbottom)
    shapes(LayerSiN).insert(wg1) 

    # Pins on the bottom waveguide side:    
    pin = Path([Point(xtop+PIN_LENGTH, ytop-w/2), Point(xtop-PIN_LENGTH, ytop-w/2)], w)
    shapes(LayerPinRecN).insert(pin)
    text = Text ("pin1", Trans(Trans.R0, xtop, ytop-w/2))
    shape = shapes(LayerPinRecN).insert(text)
    shape.text_size = 0.4/dbu      

    pin = Path([Point(xbottom-PIN_LENGTH, ytop-w/2), Point(xbottom+PIN_LENGTH, ytop-w/2)], w)
    shapes(LayerPinRecN).insert(pin)
    text = Text ("pin2", Trans(Trans.R0, xbottom, ytop-w/2))
    shape = shapes(LayerPinRecN).insert(text)
    shape.text_size = 0.4/dbu      
  
    # Pins on the left waveguide side:    
    pin = Path([Point(ytop-w/2, xtop+PIN_LENGTH), Point(ytop-w/2,xtop-PIN_LENGTH)], w)
    shapes(LayerPinRecN).insert(pin)
    text = Text ("pin3", Trans(Trans.R0,  ytop-w/2,xtop))
    shape = shapes(LayerPinRecN).insert(text)
    shape.text_size = 0.4/dbu      
    
    pin = Path([Point(ybottom+w/2, xbottom-PIN_LENGTH), Point(ybottom+w/2,xbottom+PIN_LENGTH)], w)
    shapes(LayerPinRecN).insert(pin)
    text = Text ("pin4", Trans(Trans.R0,  ybottom+w/2,xbottom))
    shape = shapes(LayerPinRecN).insert(text)
    shape.text_size = 0.4/dbu      

    #draw N
    arc_pts = arc_xy(x, y, n_w,  theta_start, theta_stop)
    arc_pts.append(pya.Point.from_dpoint(pya.DPoint(0,0))) #adding center point for polygon
    shapes(LayerNN).insert(pya.Polygon(arc_pts)) 
    
    #draw NPP
    #center
    arc_pts = arc_xy(x, y, r-npp_si-w/2,  0, 360)
    arc_pts.append(pya.Point.from_dpoint(pya.DPoint(0,0))) #adding center point for polygon    
    shapes(LayerNPPN).insert(pya.Polygon(arc_pts)) 
    #outer donut
    shapes(LayerNPPN).insert(arc_wg_xy(x, y,  r+w/2+npp_si+npp_w/2 , npp_w , theta_start, theta_stop))     

    #draw via
     #center
    arc_pts = arc_xy(x, y, vc_cw ,  0, 360)
    arc_pts.append(pya.Point.from_dpoint(pya.DPoint(0,0))) #adding center point for polygon        
    shapes(LayerVCN).insert(pya.Polygon(arc_pts))
     
    #outer donut
    shapes(LayerVCN).insert(arc_wg_xy(x, y,  4/dbu +  r + npp_w/2, vc_aw , theta_start+10, theta_stop-10))     
    
    #draw metal
    #center
    arc_pts = arc_xy(x, y, m_cw ,  0, 360)
    arc_pts.append(pya.Point.from_dpoint(pya.DPoint(0,0))) #adding center point for polygon 
    shapes(LayerMN).insert(pya.Polygon(arc_pts)) 
    #outer donut
    shapes(LayerMN).insert(arc_wg_xy(x, y,  4/dbu +  r + npp_w/2 , m_aw , theta_start, theta_stop))     

    #devrec layer
    xtop = -2/3*r
    ytop = -1*(r+g+w)
    xbottom = 4/3*r
    ybottom = ytop
    dev_array = [
        Point(ytop + 3.5*w,xbottom+npp_w),
        Point(ytop + 3.5*w,xbottom),
        Point(ytop - 3.5*w,xbottom),
        Point(ytop - 3.5*w,xtop),
        Point(ytop + 3.5*w,xtop),
        Point(xtop,ybottom+3.5*w),
        Point(xtop,ybottom-3*w),
        Point(xbottom,ybottom-3.5*w),
        Point(xbottom,ybottom+3.5*w),
        Point(xbottom+npp_w,ybottom+3.5*w),
        Point(xbottom+npp_w,xbottom+npp_w) ]

    shapes(LayerDevRecN).insert(pya.Polygon(dev_array)) 

    if self.io_wg_type == 0: # strip converter
      taper_long = to_itype(10, dbu)
      taper_short = to_itype(3, dbu)
      dev_array = [
        Point(ytop + 3.5*w,xbottom+npp_w),
        Point(ytop + 3.5*w,xbottom-taper_long),
        Point(ytop + 2.5*w,xbottom-taper_long),
        Point(ytop + w/2-overlay_ebl*2,xbottom),
        Point(ytop - w/2+overlay_ebl*2,xbottom),
        Point(ytop - 2.5*w,xbottom-taper_long),
        Point(ytop - 3.5*w,xbottom-taper_long),
        Point(ytop - 3.5*w,xtop+taper_short),
        Point(ytop - 1.5*w,xtop+taper_short),
        Point(ytop - w/2+overlay_ebl*2,xtop),
        Point(ytop + w/2-overlay_ebl*2,xtop),
        Point(ytop + 1.5*w,xtop+taper_short),
        Point(ytop + 3.5*w,xtop+taper_short),
        Point((ytop + 3.5*w + xtop)/2,(xtop+ybottom+3.5*w)/2),
        Point(xtop+taper_short,ybottom+3.5*w),
        Point(xtop+taper_short,ybottom+1.5*w),
        Point(xtop,ybottom+w/2-overlay_ebl*2),
        Point(xtop,ybottom-w/2+overlay_ebl*2),
        Point(xtop+taper_short,ybottom-1.5*w),
        Point(xtop+taper_short,ybottom-3.5*w),
        Point(xbottom-taper_long,ybottom-3.5*w),
        Point(xbottom-taper_long,ybottom-2.5*w),
        Point(xbottom,ybottom-w/2+overlay_ebl*2),
        Point(xbottom,ybottom+w/2-overlay_ebl*2),
        Point(xbottom-taper_long,ybottom+2.5*w),
        Point(xbottom-taper_long,ybottom+3.5*w),
        Point(xbottom+npp_w,ybottom+3.5*w),
        Point(xbottom+npp_w,xbottom+npp_w) ]
    # Si rib layer
    shapes(LayerSiRibN).insert(pya.Polygon(dev_array))
    
    t = Trans(Trans.R0, 0,0)
    text = Text ('Component=ebeam_irph_mrr', t)
    shape = shapes(LayerDevRecN).insert(text)
    shape.text_size = self.r*0.07/dbu
    
    # print("Done drawing the layout for - ebeam_IRPH_MRR: %.3f-%g" % ( self.r, self.g))



"""
This file is part of the SiEPIC-EBeam-PDK

Version history:

Lukas Chrostowski 2023/02/19
 - PCell takes an input of the waveguide type from a dropdown list
    Loaded from Waveguides.xml
 - Metal heater
"""

from pya import *
import pya

class wg_heater(pya.PCellDeclarationHelper):

  def __init__(self):
    # Important: initialize the super class
    super(wg_heater, self).__init__()
    
    from SiEPIC.utils import get_technology_by_name, load_Waveguides_by_Tech

    '''
    # https://github.com/KLayout/klayout/issues/879
    tech = self.layout.library().technology
    if not tech:
       tech = 'EBeam'
    self.technology_name = tech
    '''
    self.technology_name = 'EBeam'
    TECHNOLOGY = get_technology_by_name(self.technology_name)
            
    # Load all waveguides
    self.waveguide_types = load_Waveguides_by_Tech(self.technology_name)   
        
    # declare the parameters

    default_wg = 'Strip TE 1550 nm, w=500 nm'
    p = self.param("waveguide_type", self.TypeList, "Waveguide Type", default = default_wg)
#    p = self.param("waveguide_type", self.TypeList, "Waveguide Type", default = self.waveguide_types[0]['name'])

    for wa in self.waveguide_types:
        p.add_choice(wa['name'],wa['name'])
    self.param("length", self.TypeDouble, "Length", default = 100)
    self.param("path", self.TypeShape, "Path", default = DPath([DPoint(0,0), DPoint(100,0)], 0.5), hidden=True)

    self.param("mh_width", self.TypeInt, "Heater width (microns)", default=4)
    self.param("mhlayer", self.TypeLayer, "Heater Layer", default=TECHNOLOGY['M1_heater'])
    self.param("mllayer", self.TypeLayer, "Metal contact layer", default=TECHNOLOGY['M2_router'])
    self.param("metal_width", self.TypeDouble, "Metal contact width", default=10)

    self.param("wg_width", self.TypeDouble, "Waveguide width (microns)", default = 0.5, readonly=True)
    self.param("pinrec", self.TypeLayer, "PinRec Layer", default = TECHNOLOGY['PinRec'])
    self.param("devrec", self.TypeLayer, "DevRec Layer", default = TECHNOLOGY['DevRec'])
    self.param("pinrecm", self.TypeLayer, "PinRecM Layer (metal)", default=TECHNOLOGY['PinRecM'])
    self.param("text", self.TypeLayer, "Text Layer", default = LayerInfo(10, 0))
    
    self.cellName="wg_heater"

  def display_text_impl(self):
    # Provide a descriptive text for the cell
    return "%s_%s" % (self.cellName, self.length)
  
  def coerce_parameters_impl(self):
    params = [t for t in self.waveguide_types if t['name'] == self.waveguide_type]
    if not params:
        raise Exception ('Waveguides.XML not correctly configured')
    from SiEPIC.extend import to_itype
    self.wg_width = float(params[0]['width'])
    
  def can_create_from_shape_impl(self):
    return self.shape.is_path()

  def transformation_from_shape_impl(self):
    return Trans(Trans.R0,0,0)

  def parameters_from_shape_impl(self):
    self.path = self.shape.path
        
  def produce_impl(self):

    # https://github.com/KLayout/klayout/issues/879
    # tech = self.layout.library().technology
        
    # Make sure the technology name is associated with the layout
    #  PCells don't seem to know to whom they belong!
    if self.layout.technology_name == '':
        self.layout.technology_name = self.technology_name


    # fetch the parameters
    dbu = self.layout.dbu
    ly = self.layout
    shapes = self.cell.shapes

    from SiEPIC.utils import get_technology_by_name, load_Waveguides_by_Tech
    TECHNOLOGY = get_technology_by_name('EBeam')
    params = [t for t in self.waveguide_types if t['name'] == self.waveguide_type]
    layer = [wg['layer'] for wg in params[0]['component']]
    try:
        layer.remove('DevRec')
    except:
        pass
    if not layer:
        raise Exception ('Waveguides.XML not correctly configured')

    # fetch design layers
    LayermlN = self.layout.layer(self.mllayer)
    LayermhN = self.layout.layer(self.mhlayer)
    LayerPinRecN = self.layout.layer(self.pinrec)
    LayerDevRecN = self.layout.layer(self.devrec)
    LayerPinRecM = self.layout.layer(self.pinrecm)
    LayerWG = ly.layer(TECHNOLOGY[layer[0]])
    LayerTextN = ly.layer(self.text)


    from SiEPIC.extend import to_itype
    from SiEPIC.utils.layout import make_pin
    dbu = self.layout.dbu
    w = to_itype(self.wg_width,dbu)
    
    # Draw the waveguide geometry
    path = DPath([DPoint(0,0), DPoint(self.length,0)], self.wg_width)
    self.cell.shapes(LayerWG).insert(path.simple_polygon())

    # draw metal heater
    path = DPath([DPoint(0,0), DPoint(self.length,0)], self.mh_width)
    self.cell.shapes(LayermhN).insert(path.simple_polygon())

    # metal contact
    self.cell.shapes(LayermlN).insert(pya.DPath(
        [pya.DPoint(self.metal_width / 2, -self.mh_width/2), 
         pya.DPoint(self.metal_width / 2, self.mh_width/2)], self.metal_width))
    self.cell.shapes(LayermlN).insert(pya.DPath(
        [pya.DPoint(self.length-self.metal_width / 2, -self.mh_width/2), 
         pya.DPoint(self.length-self.metal_width / 2, self.mh_width/2)], self.metal_width))
    
    # metal pins
    mh_width = to_itype(self.mh_width,dbu)
    ml_width = to_itype(self.metal_width,dbu)
    make_pin(self.cell, "elec1", [self.metal_width / 2, self.mh_width/2], ml_width, LayerPinRecM, 90)
    make_pin(self.cell, "elec2", [self.length-self.metal_width / 2, self.mh_width/2], ml_width, LayerPinRecM, 90)

    # Create the pins on the waveguides
    make_pin(self.cell, "opt1", [0,0], w, LayerPinRecN, 180)
    make_pin(self.cell, "opt2", [to_itype(self.length,dbu),0], w, LayerPinRecN, 0)

    # Compact model information
    t = Trans(Trans.R0, 0, 0)
    cml = 'EBeam'
    if cml:
        text = Text ('Lumerical_INTERCONNECT_library=Design kits/%s' % cml, t)
        shape = shapes(LayerDevRecN).insert(text)
        shape.text_size = 0.1/dbu
        t = Trans(Trans.R0, 0, w*2)
    model = 'wg_heater'
    if model:
        text = Text ('Component=%s' % model, t)
        shape = shapes(LayerDevRecN).insert(text)
        shape.text_size = 0.1/dbu
        t = Trans(Trans.R0, 0, -w*2)
    text = Text \
      ('Spice_param:wg_length=%.3fu' %\
      (self.length), t )
    shape = shapes(LayerDevRecN).insert(text)
    shape.text_size = 0.1/dbu

    # Create the device recognition layer -- make it 1 * wg_width away from the waveguides.
    h = max(w*3*dbu, self.mh_width/2) 
    box1 = Box(0, -to_itype(h,dbu), to_itype(self.length,dbu), to_itype(h,dbu))
    shapes(LayerDevRecN).insert(box1)





'''
Create a layout of a paperclip spiral,
uses: delay line, waveguide loss cutback structure

- this can be executed directly, or loaded by the technology library
- you can resize the PCell by using the Partial select tool "L", and moving 
   the GUIDING SHAPE Box
- Waveguide types are loaded from the XML file
- Pitch of the waveguides is determined by the DevRec layer
- When executed directly, the unit test instantiates all the waveguide types

by Lukas Chrostowski, 2023
'''


import pya
from pya import *
from SiEPIC.utils import get_technology_by_name

class spiral_paperclip(pya.PCellDeclarationHelper):

    def __init__(self):
        # Important: initialize the super class
        self.cellName=self.__class__.__name__
        super(eval(self.cellName), self).__init__()
        
        self.technology_name = 'EBeam' 

        from SiEPIC.utils import get_technology_by_name, load_Waveguides_by_Tech
        self.TECHNOLOGY = get_technology_by_name(self.technology_name)
                        
        # Load all strip waveguides
        self.waveguide_types = load_Waveguides_by_Tech(self.technology_name)   
            
        # declare the parameters
        p = self.param("waveguide_type", self.TypeList, "Waveguide Type", default = self.waveguide_types[0]['name'])
        for wa in self.waveguide_types:
            p.add_choice(wa['name'],wa['name'])
        
        self.param("loops", self.TypeInt, "Number of loops", default = 2)
        self.minlength = 2*float(self.waveguide_types[0]['radius'])
        self.param("length", self.TypeDouble, "Inner length (min 2 x bend radius)", default = self.minlength)
        self.param("box", self.TypeShape, "Box", default = pya.DBox(-self.minlength, -self.minlength/2, self.minlength, self.minlength/2))
        self.param("length1", self.TypeDouble, "length: for PCell", default = self.minlength, hidden = True)
        self.param("flatten", self.TypeBoolean, "Flatten the PCell, for scripting", default = False)
        
    def display_text_impl(self):
        # Provide a descriptive text for the cell
        return "%s_%s_%s_%s" % (self.cellName, self.loops, self.length, self.waveguide_type.replace(' ','_'))

    def get_waveguide_parameters(self):
        '''
        Get the Waveguide Type from the PCell, and update waveguide parameters
        '''
        self.waveguide_params = [t for t in self.waveguide_types if t['name'] == self.waveguide_type]
        if not self.waveguide_params:
            raise Exception ('Waveguides.XML not correctly configured')
        self.waveguide_params = self.waveguide_params [0]
        wg_params = self.waveguide_params
        wg_params2 = wg_params
        
        # check for compound type
        if 'compound_waveguide' in wg_params:
            sm_wg = self.waveguide_params['compound_waveguide']['singlemode']
            mm_wg = self.waveguide_params['compound_waveguide']['multimode']
            wg_params = [t for t in self.waveguide_types if t['name'] == sm_wg][0]
            wg_params2 = [t for t in self.waveguide_types if t['name'] == mm_wg][0]

        # DevRec width
        if ('DevRec' not in [wg['layer'] for wg in wg_params2['component']]):
            from SiEPIC import _globals
            self.devrec = max([float(wg['width']) for wg in wg_params2['component']]) + _globals.WG_DEVREC_SPACE * 2
        else:
            self.devrec = float([f for f in wg_params2['component'] if f['layer']=='DevRec'][0]['width'])

        # Radius
        self.radius = float(wg_params['radius'])

        # Waveguide pin width
        self.wg_width = float(wg_params['width'])

    def coerce_parameters_impl(self):
        self.get_waveguide_parameters()
        
        # We employ coerce_parameters_impl to decide whether the handle or the 
        # numeric parameters have changed. We update the 
        # numerical value or the shape, depending on which on has not changed.
        w = None
        if isinstance(self.box, pya.Box): 
            w = self.box.right*self.layout.dbu
            # print('%s Box: width %s ' %(self.cellName,w))
        if isinstance(self.box, pya.DBox): 
            w = self.box.right
            # print('%s DBox: width %s ' %(self.cellName,w))
        if w != None and abs(self.length1-self.length) < 1e-6:
            if w < self.minlength:
                w = self.minlength
            # print('%s update from GUI Box: %s ' %(self.cellName,w))
            self.length = w
            self.length1 = w
        else:
            # print('%s update from PCell parameters: %s ' %(self.cellName,self.length1))
            self.length1 = self.length
        self.box = pya.DBox(-self.length,-self.minlength/2,self.length,self.minlength/2)
                    

    def can_create_from_shape_impl(self):
        return self.shape.is_box()

    def transformation_from_shape_impl(self):
        # Implement the "Create PCell from shape" protocol: we use the center of the shape's
        # bounding box to determine the transformation
        return pya.Trans(self.shape.bbox().center())

    def parameters_from_shape_impl(self):
        # self.path = self.shape.path
        # Implement the "Create PCell from shape" protocol: we set r and l from the shape's 
        # bounding box width and layer
        self.width = self.shape.bbox().width() * self.layout.dbu
        self.height = self.shape.bbox().height() * self.layout.dbu
    
    def produce_impl(self):
    
        self.get_waveguide_parameters()
        devrec=self.devrec
        radius = self.radius

        # SBend offset in the middle of the spiral
        if 'sbends' in self.waveguide_params:
            # Use the Bezier S-Bend (more space efficient)
            # if sbends=True in Waveguide.xml
            offset = radius - devrec/2  
            extra = 0
        else:
            # Use 2 x 90 degree bends
            offset = radius
            extra = devrec
    
        # Ensure the length is sufficient
        self.minlength = 2*radius
        length0 = max(self.minlength, self.length)
        
        # min loops
        self.loops = max(self.loops, 1)
    
        # spiral points    
        points = [DPoint(-length0,offset), DPoint(0.0,offset), DPoint(0.0,-offset), DPoint(length0,-offset)]
        for i in range(1,self.loops*2,2):
            points.insert(0, DPoint(-length0-devrec*(i-1),offset-radius*2-devrec*(i-1)-extra))
            points.insert(0, DPoint(length0+devrec*i,offset-radius*2-devrec*(i-1)-extra))
            points.insert(0, DPoint(length0+devrec*i,-offset+radius*2+devrec*i+extra))
            points.insert(0, DPoint(-length0-devrec*(i+1),-offset+radius*2+devrec*i+extra))
            points.append(DPoint(length0+devrec*(i-1),radius*2-offset+devrec*(i-1)+extra))
            points.append(DPoint(-length0-devrec*i,radius*2-offset+devrec*(i-1)+extra))
            points.append(DPoint(-length0-devrec*i,-radius*2+offset-devrec*i-extra))
            points.append(DPoint(length0+devrec*(i+1),-radius*2+offset-devrec*i-extra))
        points.append(DPoint(length0+devrec*(i+1),radius*2-offset+devrec*(i+1)+extra))
        points.append(DPoint(-length0-devrec*(i+1),radius*2-offset+devrec*(i+1)+extra))
        points.pop(0)
        points.insert(0, DPoint(-length0-devrec*(i+1),-offset+radius*2+devrec*i+extra))

        # Create a path and waveguide    
        path  = DPath(points, 0.5)
        self.layout.technology_name = self.technology_name # required otherwise "create_cell" doesn't load
        pcell = self.layout.create_cell("Waveguide", self.technology_name, 
                    {"path": path,
                     "waveguide_type": self.waveguide_type})
        t = Trans(Trans.R0, 0, 0)
        self.cell.insert(CellInstArray(pcell.cell_index(), t))

        # If the waveguide is a Compound type:
        # Flatten the PCell and remove the DevRec and PinRec layers
        if 'compound_waveguide' in self.waveguide_params and self.flatten:
            self.cell.flatten(True)
            LayerPinRecN = self.layout.layer(self.TECHNOLOGY['PinRec'])
            LayerDevRecN = self.layout.layer(self.TECHNOLOGY['DevRec']) 
            self.cell.clear(LayerPinRecN)
            bbox = self.cell.bbox()
            self.cell.clear(LayerDevRecN)     
            self.cell.clear(self.layout.layer(self.TECHNOLOGY['Waveguide']))
            self.cell.shapes(LayerDevRecN).insert(bbox)
    
            # Create the pins on the input & output waveguides
            from SiEPIC.utils.layout import make_pin
            make_pin(self.cell, "optA", [-length0-devrec*(i+1),radius*2-offset+devrec*(i+1)+extra], self.wg_width, LayerPinRecN, 180)
            make_pin(self.cell, "optB", [-length0-devrec*(i+1),-offset+radius*2+devrec*i+extra], self.wg_width, LayerPinRecN, 180)


# Save the path, used for loading WAVEGUIDES.XML
import os
filepath = os.path.dirname(os.path.realpath(__file__))


class test_lib(Library):
  def __init__(self):  
    tech = 'EBeam'
    library = tech + 'test_lib'
    self.technology=tech
    self.layout().register_pcell('spiral_paperclip', spiral_paperclip())
    self.register(library)
 
if __name__ == "__main__":
    print('Test layout for: Spiral Paperclip')
    
    from SiEPIC.utils.layout import new_layout, floorplan
    from SiEPIC.scripts import zoom_out

    # load the test library, and technology
    t = test_lib()
    tech = t.technology

    # Create a new layout for the chip floor plan   
    topcell, ly = new_layout(tech, "test", GUI = True, overwrite = True)
    #floorplan(topcell, 100e3, 100e3)
        
    # instantiate the cell
    library = tech + 'test_lib'

    if 0:
        # default waveguide
        pcell = ly.create_cell("spiral_paperclip", library, {})
        t = Trans(Trans.R0, 0, 0)
        topcell.insert(CellInstArray(pcell.cell_index(), t))
    
        # multimode
        pcell = ly.create_cell("spiral_paperclip", library, {
                'waveguide_type':'Si routing TE 1550 nm (compound waveguide)',
                'length':100,
                'loops':1})
        t = Trans(Trans.R0, 0, 40000)
        topcell.insert(CellInstArray(pcell.cell_index(), t))

    # Create spirals for all the types of waveguides
    from SiEPIC.utils import load_Waveguides_by_Tech
    waveguide_types = load_Waveguides_by_Tech(tech)   
    y = 0
    for wg in waveguide_types:
        pcell = ly.create_cell("spiral_paperclip", library, {
                'waveguide_type':wg['name'],
                'length':100,
                'loops':1})
        t = Trans(Trans.R0, 0, y - pcell.bbox().bottom)
        inst = topcell.insert(CellInstArray(pcell.cell_index(), t))
        y += pcell.bbox().height()+2000
    
    zoom_out(topcell)



import pya
from pya import *
from SiEPIC.utils import get_technology_by_name

class ebeam_bragg_apodized(pya.PCellDeclarationHelper):
  """
  Author:   Mustafa Hammood 
            Mustafa@siepic.com
  """

  def __init__(self):

    # Important: initialize the super class
    super(ebeam_bragg_apodized, self).__init__()
    TECHNOLOGY = get_technology_by_name('EBeam')

    # declare the parameters
    self.param("number_of_periods", self.TypeInt, "Number of grating periods", default = 300)     
    self.param("grating_period", self.TypeDouble, "Grating period (microns)", default = 0.317)     
    self.param("corrugation_width", self.TypeDouble, "Corrugration width (microns)", default = 0.05)     
    self.param("misalignment", self.TypeDouble, "Grating misalignment (microns)", default = 0.0)     
    self.param("sinusoidal", self.TypeBoolean, "Grating Type (Rectangular=False, Sinusoidal=True)", default = False)     
    self.param("wg_width", self.TypeDouble, "Waveguide width", default = 0.5)     
    self.param("index", self.TypeDouble, "Gaussian Index", default = 3)
    self.param("layer", self.TypeLayer, "Layer", default = TECHNOLOGY['Si'])
    self.param("pinrec", self.TypeLayer, "PinRec Layer", default = TECHNOLOGY['PinRec'])
    self.param("devrec", self.TypeLayer, "DevRec Layer", default = TECHNOLOGY['DevRec'])
#    self.param("textl", self.TypeLayer, "Text Layer", default = LayerInfo(10, 0))

  def display_text_impl(self):
    # Provide a descriptive text for the cell
    return "ebeam_bragg_apodized_%s-%.3f-%.3f-%.3f" % \
    (self.number_of_periods, self.grating_period, self.corrugation_width, self.misalignment)
  
  def coerce_parameters_impl(self):
    pass

  def can_create_from_shape(self, layout, shape, layer):
    return False
    
  def produce_impl(self):
    import math

    # fetch the parameters
    dbu = self.layout.dbu
    ly = self.layout
    shapes = self.cell.shapes

    LayerSi = self.layer
    LayerSiN = ly.layer(LayerSi)
    LayerPinRecN = ly.layer(self.pinrec)
    LayerDevRecN = ly.layer(self.devrec)

    from SiEPIC.extend import to_itype
    
    # Draw the Bragg grating:
    box_width = int(round(self.grating_period/2/dbu))
    grating_period = int(round(self.grating_period/dbu))
    w = to_itype(self.wg_width,dbu)
    GaussianIndex = self.index
    half_w = w/2
    half_corrugation_w = int(round(self.corrugation_width/2/dbu))
    misalignment = int(round(self.misalignment/dbu))
    N = self.number_of_periods
    if self.sinusoidal:
      npoints_sin = 40
      for i in range(0,self.number_of_periods):
        x = (round((i * self.grating_period)/dbu))
        profileFunction = math.exp( -0.5*(2*GaussianIndex*(i-N/2)/(N))**2 )
        profile = int(round(self.corrugation_width/2/dbu))*profileFunction;
        box1 = Box(x, 0, x + box_width, half_w+profile)
        pts1 = [Point(x,0)]
        pts3 = [Point(x + misalignment,0)]
        for i1 in range(0,npoints_sin+1):
          x1 = i1 * 2* math.pi / npoints_sin
          y1 = round(profile*math.sin(x1))
          x1 = round(x1/2/math.pi*grating_period)
#          print("x: %s, y: %s" % (x1,y1))
          pts1.append( Point(x + x1,half_w+y1 ) )
          pts3.append( Point(x + misalignment + x1,-half_w-y1 ) )
        pts1.append( Point(x + grating_period, 0) )
        pts3.append( Point(x + grating_period + misalignment, 0) )
        shapes(LayerSiN).insert(Polygon(pts1))
        shapes(LayerSiN).insert(Polygon(pts3))
      length = x + grating_period + misalignment
      if misalignment > 0:
        # extra piece at the end:
        box2 = Box(x + grating_period, 0, length, half_w)
        shapes(LayerSiN).insert(box2)
        # extra piece at the beginning:
        box3 = Box(0, 0, misalignment, -half_w)
        shapes(LayerSiN).insert(box3)

    else:
    
      
      
      for i in range(0,self.number_of_periods):
        x = int(round((i * self.grating_period)/dbu))
        
        profileFunction = math.exp( -0.5*(2*GaussianIndex*(i-N/2)/(N))**2 )
        profile = int(round(self.corrugation_width/2/dbu))*profileFunction;
        box1 = Box(x, 0, x + box_width, half_w+profile)
        box2 = Box(x + box_width, 0, x + grating_period, half_w-profile)
        box3 = Box(x + misalignment, 0, x + box_width + misalignment, -half_w-profile)
        box4 = Box(x + box_width + misalignment, 0, x + grating_period + misalignment, -half_w+profile)
        shapes(LayerSiN).insert(box1)
        shapes(LayerSiN).insert(box2)
        shapes(LayerSiN).insert(box3)
        shapes(LayerSiN).insert(box4)
      length = x + grating_period + misalignment
      if misalignment > 0:
        # extra piece at the end:
        box2 = Box(x + grating_period, 0, length, half_w)
        shapes(LayerSiN).insert(box2)
        # extra piece at the beginning:
        box3 = Box(0, 0, misalignment, -half_w)
        shapes(LayerSiN).insert(box3)

    
    # Create the pins on the waveguides, as short paths:
    from SiEPIC._globals import PIN_LENGTH as pin_length

    t = Trans(Trans.R0, 0,0)
    pin = Path([Point(pin_length/2, 0), Point(-pin_length/2, 0)], w)
    pin_t = pin.transformed(t)
    shapes(LayerPinRecN).insert(pin_t)
    text = Text ("pin1", t)
    shape = shapes(LayerPinRecN).insert(text)
    shape.text_size = 0.4/dbu

    t = Trans(Trans.R0, length,0)
    pin = Path([Point(-pin_length/2, 0), Point(pin_length/2, 0)], w)
    pin_t = pin.transformed(t)
    shapes(LayerPinRecN).insert(pin_t)
    text = Text ("pin2", t)
    shape = shapes(LayerPinRecN).insert(text)
    shape.text_size = 0.4/dbu

    # Compact model information
    t = Trans(Trans.R0, 0, 0)
    text = Text ('Lumerical_INTERCONNECT_library=Design kits/ebeam', t)
    shape = shapes(LayerDevRecN).insert(text)
    shape.text_size = 0.1/dbu
    t = Trans(Trans.R0, length/10, 0)
    text = Text ('Component=ebeam_bragg_te1550', t)
    shape = shapes(LayerDevRecN).insert(text)
    shape.text_size = 0.1/dbu
    t = Trans(Trans.R0, length/9, 0)
    text = Text \
      ('Spice_param:number_of_periods=%s grating_period=%.3fu corrugation_width=%.3fu misalignment=%.3fu sinusoidal=%s' %\
      (self.number_of_periods, self.grating_period, self.corrugation_width, self.misalignment, int(self.sinusoidal)), t )
    shape = shapes(LayerDevRecN).insert(text)
    shape.text_size = 0.1/dbu

    # Create the device recognition layer -- make it 1 * wg_width away from the waveguides.
    t = Trans(Trans.R0, 0,0)
    path = Path([Point(0, 0), Point(length, 0)], 3*w)
    shapes(LayerDevRecN).insert(path.simple_polygon())
